from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner


class MissingVitaTitleIDError(MissingGameExecutableError):
    """Raise when the Title ID field has not be supplied to the Vita runner game options"""

    def __init__(self, message=None, *args, **kwargs):
        if not message:
            message = _("The Vita App has no Title ID set")

        super().__init__(message, *args, **kwargs)


class vita3k(Runner):
    human_name = _("Vita3K")
    platforms = [_("Sony PlayStation Vita")]
    description = _("Sony PlayStation Vita emulator")
    runnable_alone = True
    runner_executable = "vita3k/Vita3K-x86_64.AppImage"
    flatpak_id = None
    download_url = "https://github.com/Vita3K/Vita3K/releases/download/continuous/Vita3K-x86_64.AppImage"
    game_options = [
        {
            "option": "main_file",
            "type": "string",
            "label": _("Title ID of Installed Application"),
            "argument": "-r",
            "help": _(
                'Title ID of installed application. Eg."PCSG00042". User installed apps are located in '
                "ux0:/app/&lt;title-id&gt;."
            ),
        }
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": True,
            "argument": "-F",
            "help": _("Start the emulator in fullscreen mode."),
        },
        {
            "option": "config",
            "type": "file",
            "label": _("Config location"),
            "argument": "-c",
            "help": _(
                'Get a configuration file from a given location. If a filename is given, it must end with ".yml", '
                "otherwise it will be assumed to be a directory."
            ),
        },
        {
            "option": "load-config",
            "label": _("Load configuration file"),
            "type": "bool",
            "argument": "-f",
            "help": _('If trues, informs the emualtor to load the config file from the "Config location" option.'),
        },
    ]

    # Vita3k uses an AppImage and doesn't require the Lutris runtime.
    system_options_override = [{"option": "disable_runtime", "default": True}]

    def play(self):
        """Run the game."""
        arguments = self.get_command()

        # adds arguments from the supplied option dictionary to the arguments list
        def append_args(option_dict, config):
            for option in option_dict:
                if option["option"] not in config:
                    continue
                if option["type"] == "bool":
                    if self.runner_config.get(option["option"]):
                        if "argument" in option:
                            arguments.append(option["argument"])
                elif option["type"] == "choice":
                    if self.runner_config.get(option["option"]) != "off":
                        if "argument" in option:
                            arguments.append(option["argument"])
                        arguments.append(config.get(option["option"]))
                elif option["type"] in ("string", "file"):
                    if "argument" in option:
                        arguments.append(option["argument"])
                    arguments.append(config.get(option["option"]))
                else:
                    raise RuntimeError("Unhandled type %s" % option["type"])

        # Append the runner arguments first, and game arguments afterwards
        append_args(self.runner_options, self.runner_config)

        title_id = self.game_config.get("main_file") or ""
        if not title_id:
            raise MissingVitaTitleIDError(_("The Vita App has no Title ID set"))

        append_args(self.game_options, self.game_config)
        return {"command": arguments}

    @property
    def game_path(self):
        return self.game_config.get(self.entry_point_option, "")



import abc
import re
from collections import defaultdict
from itertools import repeat
from typing import Any, Dict, List, Set, Union

from lutris import settings
from lutris.database import games as games_db
from lutris.database import sql
from lutris.gui.widgets import NotificationSource
from lutris.util.strings import get_natural_sort_key

CATEGORIES_UPDATED = NotificationSource()


class _SmartCategory(abc.ABC):
    """Abstract class to define smart categories. Smart categories are automatically defined based on a rule."""

    @abc.abstractmethod
    def get_name(self) -> str:
        pass

    def get_game_ids(self) -> Set[str]:
        return set(game["id"] for game in self.get_games())

    @abc.abstractmethod
    def get_games(self) -> List[Any]:
        pass


class _SmartUncategorizedCategory(_SmartCategory):
    """A SmartCategory that resolves to all uncategorized games."""

    def get_name(self) -> str:
        return ".uncategorized"

    def get_game_ids(self) -> Set[str]:
        return get_uncategorized_game_ids()

    def get_games(self) -> List[Any]:
        return get_uncategorized_games()


# All smart categories should be added to this variable.
# TODO: Expose a way for the users to define new smart categories.
_SMART_CATEGORIES: List[_SmartCategory] = [_SmartUncategorizedCategory()]


def strip_category_name(name):
    """ "This strips the name given, and also removes extra internal whitespace."""
    name = (name or "").strip()
    if not is_reserved_category(name):
        name = re.sub(" +", " ", name)  # Remove excessive whitespaces
    return name


def is_reserved_category(name):
    """True if name is None, blank or is a name Lutris uses internally, or
    starts with '.' for future expansion."""
    return not name or name[0] == "." or name in ["all", "favorite"]


def get_categories() -> List[Dict[str, Union[int, str]]]:
    """Get the list of every category in database."""
    # Categories look like [{"id": 1, "name": "My Category"}, ...]
    return sql.db_select(settings.DB_PATH, "categories")


def get_all_games_categories():
    games_categories = defaultdict(list)
    for row in sql.db_select(settings.DB_PATH, "games_categories"):
        games_categories[row["game_id"]].append(row["category_id"])
    return games_categories


def get_category_by_name(name):
    """Return a category by name"""
    categories = sql.db_select(settings.DB_PATH, "categories", condition=("name", name))
    if categories:
        return categories[0]


def get_category_by_id(category_id):
    """Return a category by name"""
    categories = sql.db_select(settings.DB_PATH, "categories", condition=("id", category_id))
    if categories:
        return categories[0]


def normalized_category_names(name: str, subname_allowed: bool = False) -> List[str]:
    """Searches for a category name case-insensitively and returns all matching names;
    if none match, it just returns 'name' as is.

    If subname_allowed is true but name is not a match for any category, we'll look for
    any category that contains the name as a substring instead before falling back to
    'name' itself."""
    query = "SELECT name FROM categories WHERE name=? COLLATE NOCASE"
    parameters = (name,)
    names = [cat["name"] for cat in sql.db_query(settings.DB_PATH, query, parameters)]

    if not names and subname_allowed:
        query = "SELECT name FROM categories WHERE name LIKE ? COLLATE NOCASE"
        parameters = (f"%{name}%",)
        names = [cat["name"] for cat in sql.db_query(settings.DB_PATH, query, parameters)]

    return names or [name]


def get_game_ids_for_categories(included_category_names=None, excluded_category_names=None):
    """Get the ids of games in database."""
    filters = []
    parameters = []

    if included_category_names:
        # Query that finds games in the included categories
        query = (
            "SELECT games.id FROM games "
            "INNER JOIN games_categories ON games.id = games_categories.game_id "
            "INNER JOIN categories ON categories.id = games_categories.category_id"
        )
        filters.append("categories.name IN (%s)" % ", ".join(repeat("?", len(included_category_names))))
        parameters.extend(included_category_names)
    else:
        # Or, if you listed none, we fall back to all games
        query = "SELECT games.id FROM games"

    if excluded_category_names:
        # Sub-query to exclude the excluded categories, if any.
        exclude_filter = (
            "NOT EXISTS(SELECT * FROM games_categories AS gc "
            "INNER JOIN categories AS c ON gc.category_id = c.id "
            "WHERE gc.game_id = games.id "
            "AND c.name IN (%s))" % ", ".join(repeat("?", len(excluded_category_names)))
        )
        filters.append(exclude_filter)
        parameters.extend(excluded_category_names)

    if filters:
        query += " WHERE %s" % " AND ".join(filters)

    result = set(game["id"] for game in sql.db_query(settings.DB_PATH, query, tuple(parameters)))
    for smart_cat in _SMART_CATEGORIES:
        if excluded_category_names is not None and smart_cat.get_name() in excluded_category_names:
            continue
        if included_category_names is not None and smart_cat.get_name() not in included_category_names:
            continue
        result |= smart_cat.get_game_ids()

    return list(sorted(result))


def get_uncategorized_game_ids() -> Set[str]:
    """Returns the ids of games that are in no categories. We do not count
    the 'favorites' category, but we do count '.hidden'- hidden games are hidden
    from this too."""
    query = (
        "SELECT games.id FROM games WHERE NOT EXISTS("
        "SELECT * FROM games_categories "
        "INNER JOIN categories ON categories.id = games_categories.category_id "
        "AND categories.name NOT IN ('all', 'favorite') "
        "WHERE games.id = games_categories.game_id)"
    )
    uncategorized = sql.db_query(settings.DB_PATH, query)
    return set(row["id"] for row in uncategorized)


def get_uncategorized_games() -> List[Any]:
    """Return a list of currently running games"""
    games = games_db.get_games_by_ids(get_uncategorized_game_ids())

    def get_key(g: Dict[str, Any]):
        """Sort in the default order for Lutris- installed games first, then by name."""
        name = str(g.get("name") or "")
        installed = bool(g.get("installed"))
        return not installed, get_natural_sort_key(name)

    games.sort(key=get_key)
    return games


def get_categories_in_game(game_id):
    """Get the categories of a game in database."""
    query = (
        "SELECT categories.name FROM categories "
        "JOIN games_categories ON categories.id = games_categories.category_id "
        "JOIN games ON games.id = games_categories.game_id "
        "WHERE games.id=?"
    )
    return [category["name"] for category in sql.db_query(settings.DB_PATH, query, (game_id,))]


def add_category(category_name, no_signal: bool = False):
    """Add a category to the database"""
    cat = sql.db_insert(settings.DB_PATH, "categories", {"name": category_name})
    if not no_signal:
        CATEGORIES_UPDATED.fire()
    return cat


def redefine_category(category_id: int, new_name: str, no_signal: bool = False) -> None:
    query = "UPDATE categories SET name=? WHERE id=?"

    with sql.db_cursor(settings.DB_PATH) as cursor:
        sql.cursor_execute(cursor, query, (new_name, category_id))
    if not no_signal:
        CATEGORIES_UPDATED.fire()


def remove_category(category_id: int, no_signal: bool = False) -> None:
    queries = ["DELETE FROM games_categories WHERE category_id=?", "DELETE FROM categories WHERE id=?"]

    for query in queries:
        with sql.db_cursor(settings.DB_PATH) as cursor:
            sql.cursor_execute(cursor, query, (category_id,))

    if not no_signal:
        CATEGORIES_UPDATED.fire()


def add_game_to_category(game_id, category_id):
    """Add a category to a game"""
    return sql.db_insert(settings.DB_PATH, "games_categories", {"game_id": game_id, "category_id": category_id})


def remove_category_from_game(game_id, category_id):
    """Remove a category from a game"""
    query = "DELETE FROM games_categories WHERE category_id=? AND game_id=?"
    with sql.db_cursor(settings.DB_PATH) as cursor:
        sql.cursor_execute(cursor, query, (category_id, game_id))


def remove_unused_categories():
    """Remove all categories that have no games associated with them"""

    delete_orphaned_games = (
        "DELETE FROM games_categories "
        "WHERE NOT EXISTS(SELECT * FROM games WHERE game_id=id) "
        "OR NOT EXISTS(SELECT * FROM categories WHERE category_id=id)"
    )

    with sql.db_cursor(settings.DB_PATH) as cursor:
        sql.cursor_execute(cursor, delete_orphaned_games, ())

    find_orphaned_categories = (
        "SELECT categories.* FROM categories "
        "LEFT JOIN games_categories ON categories.id = games_categories.category_id "
        "WHERE games_categories.category_id IS NULL"
    )

    empty_categories = sql.db_query(settings.DB_PATH, find_orphaned_categories)
    for category in empty_categories:
        if category["name"] == "favorite":
            continue

        delete_orphaned_categories = "DELETE FROM categories WHERE categories.id=?"
        with sql.db_cursor(settings.DB_PATH) as cursor:
            sql.cursor_execute(cursor, delete_orphaned_categories, (category["id"],))



# Third Party Libraries
from gi.repository import Gtk


class BaseApplicationWindow(Gtk.ApplicationWindow):
    """Window used to guide the user through a issue reporting process"""

    def __init__(self, application):
        Gtk.ApplicationWindow.__init__(self, icon_name="net.lutris.Lutris", application=application)
        self.application = application
        self.set_show_menubar(False)

        self.set_position(Gtk.WindowPosition.CENTER)

        self.vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12, visible=True)
        self.vbox.set_margin_top(18)
        self.vbox.set_margin_bottom(18)
        self.vbox.set_margin_right(18)
        self.vbox.set_margin_left(18)
        self.add(self.vbox)
        self.action_buttons = Gtk.Box(spacing=6)
        self.vbox.pack_end(self.action_buttons, False, False, 0)

    def get_action_button(self, label, handler=None, tooltip=None):
        """Returns a button that can be used for the action bar"""
        button = Gtk.Button.new_with_mnemonic(label)
        if handler:
            button.connect("clicked", handler)
        if tooltip:
            button.set_tooltip_text(tooltip)
        return button

    def present(self):  # pylint: disable=arguments-differ
        """The base implementation doesn't always work, this one does."""
        self.set_keep_above(True)
        super().present()
        self.set_keep_above(False)
        super().present()



import os

from lutris.util import system
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.nvidia import get_nvidia_dll_path
from lutris.util.wine.dll_manager import DLLManager


class DXVKNVAPIManager(DLLManager):
    name = "dxvk-nvapi"
    human_name = "DXVK-NVAPI"
    # apparently, nvofapi.dll (the 32 bit version) is not being included here -
    # see https://github.com/jp7677/dxvk-nvapi/pull/213
    managed_dlls = ("nvapi", "nvapi64", "nvml", "nvofapi64")
    releases_url = "https://api.github.com/repos/lutris/dxvk-nvapi/releases"
    dlss_dlls = ("nvngx", "_nvngx")

    def can_enable(self):
        return LINUX_SYSTEM.is_vulkan_supported()

    def disable_dll(self, system_dir, _arch, dll):  # pylint: disable=unused-argument
        """Remove DLL from Wine prefix"""
        wine_dll_path = os.path.join(system_dir, "%s.dll" % dll)
        if system.path_exists(wine_dll_path):
            os.remove(wine_dll_path)

    def enable(self):
        """Enable Dlls for the current prefix"""
        super().enable()
        dlss_dll_dir = get_nvidia_dll_path()
        if not dlss_dll_dir:
            return

        windows_path = os.path.join(self.prefix, "drive_c/windows")
        system_dir = os.path.join(windows_path, "system32")
        for dll in self.dlss_dlls:
            dll_path = os.path.join(dlss_dll_dir, "%s.dll" % dll)
            self.enable_dll(system_dir, "x64", dll_path)

    def disable(self):
        """Disable DLLs for the current prefix"""
        super().disable()
        windows_path = os.path.join(self.prefix, "drive_c/windows")
        system_dir = os.path.join(windows_path, "system32")
        for dll in self.dlss_dlls:
            self.disable_dll(system_dir, "x64", dll)



"""Manipulates installer files"""

import os
from gettext import gettext as _
from typing import Optional
from urllib.parse import urlparse

from lutris.cache import get_url_cache_path, has_valid_custom_cache_path, save_to_cache
from lutris.gui.widgets.download_progress_box import DownloadProgressBox
from lutris.installer.errors import ScriptingError
from lutris.util import system
from lutris.util.downloader import Downloader
from lutris.util.log import logger
from lutris.util.strings import gtk_safe_urls


class InstallerFile:
    """Representation of a file in the `files` sections of an installer"""

    def __init__(self, game_slug, file_id, file_meta):
        self.game_slug = game_slug
        self.id = file_id.replace("-", "_")  # pylint: disable=invalid-name
        self._file_meta = file_meta
        self._dest_file_override = None  # Used to override the destination
        self._dest_file_found = None  # Lazy storage for the resolved destination file
        if isinstance(self._file_meta, dict):
            self._downloader = self._file_meta.get("downloader")
        else:
            self._downloader = None

    def copy(self):
        """Copies this file object, so the copy can be modified safely."""
        _file_meta = self._file_meta
        if isinstance(_file_meta, dict):
            _file_meta = _file_meta.copy()

        file = InstallerFile(self.game_slug, self.id, _file_meta)
        file._dest_file_override = self._dest_file_override
        file._dest_file_found = self._dest_file_found
        file._downloader = self._downloader
        return file

    @property
    def url(self):
        _url = ""
        if isinstance(self._file_meta, dict):
            if "url" not in self._file_meta:
                raise ScriptingError(_("missing field `url` for file `%s`") % self.id)
            _url = self._file_meta["url"]
        else:
            _url = self._file_meta
        if _url.startswith("/"):
            return "file://" + _url
        return _url

    def set_url(self, url):
        """Change the internal value of the URL"""
        if isinstance(self._file_meta, dict):
            self._file_meta["url"] = url
        else:
            self._file_meta = url

    @property
    def filename(self):
        if isinstance(self._file_meta, dict):
            if "filename" not in self._file_meta:
                raise ScriptingError(_("missing field `filename` in file `%s`") % self.id)
            return self._file_meta["filename"]
        if self._file_meta.startswith("N/A"):
            if self.uses_pga_cache() and os.path.isdir(self._cache_path):
                return self.cached_filename
            return ""
        if self.url.startswith("$STEAM"):
            return self.url
        return os.path.basename(self._file_meta)

    def get_alternate_filenames(self):
        if isinstance(self._file_meta, dict):
            return self._file_meta.get("alternate_filenames") or []
        return []

    @property
    def referer(self):
        if isinstance(self._file_meta, dict):
            return self._file_meta.get("referer")

    @property
    def downloader(self) -> Optional[Downloader]:
        if callable(self._downloader):
            self._downloader = self._downloader(self)

        return self._downloader

    @property
    def checksum(self):
        if isinstance(self._file_meta, dict):
            return self._file_meta.get("checksum")

    @property
    def dest_file(self):
        def find_dest_file():
            for alt_name in self.get_alternate_filenames():
                alt_path = os.path.join(self._cache_path, alt_name)
                if os.path.isfile(alt_path):
                    return alt_path

            return os.path.join(self._cache_path, self.filename)

        if self._dest_file_override:
            return self._dest_file_override

        if not self._dest_file_found:
            self._dest_file_found = find_dest_file()

        return self._dest_file_found

    @dest_file.setter
    def dest_file(self, value):
        self._dest_file_override = value

    @property
    def download_file(self):
        """This is the actual path to download to; this file is renamed to the
        dest_file when complete."""
        return self.dest_file + ".tmp"

    def override_dest_file(self, new_dest_file):
        """Called by the UI when the user selects a file path."""
        self.dest_file = new_dest_file

    @property
    def is_dest_file_overridden(self):
        return bool(self._dest_file_override)

    def get_dest_files_by_id(self):
        return {self.id: self.dest_file}

    def __str__(self):
        return "%s/%s" % (self.game_slug, self.id)

    @property
    def auxiliary_info(self):
        """Provides a small bit of additional descriptive texts to show in the UI."""
        return None

    @property
    def human_url(self):
        """Return the url in human-readable format"""
        if self.url.startswith("N/A"):
            # Ask the user where the file is located
            parts = self.url.split(":", 1)
            if len(parts) == 2:
                return parts[1]
            return "Please select file '%s'" % self.id
        return self.url

    def get_label(self):
        """Return a human readable label for installer files"""
        url = self.url
        if url.startswith("http"):
            parsed = urlparse(url)
            label = _("{file} on {host}").format(file=self.filename, host=parsed.netloc)
        elif url.startswith("N/A"):
            label = url[3:].lstrip(":")
        else:
            label = url
        return gtk_safe_urls(label)

    @property
    def cached_filename(self):
        """Return the filename of the first file in the cache path"""
        cache_files = os.listdir(self._cache_path)
        if cache_files:
            return cache_files[0]
        return ""

    @property
    def default_provider(self):
        """Return file provider used"""
        if self.url.startswith("$STEAM"):
            return "steam"
        if self.is_cached:
            return "pga"
        if self.url.startswith("N/A"):
            return "user"
        if self.is_downloadable():
            return "download"
        raise ValueError("Unsupported provider for %s" % self.url)

    @property
    def providers(self):
        """Return all supported providers"""
        _providers = set()
        if self.url.startswith("$STEAM"):
            _providers.add("steam")
        if self.is_cached:
            _providers.add("pga")
        if self.url.startswith("N/A"):
            _providers.add("user")
        if self.is_downloadable():
            _providers.add("download")
        return _providers

    def is_downloadable(self):
        """Return True if the file can be downloaded (even from the local filesystem)"""
        return self.url.startswith(("http", "file"))

    def uses_pga_cache(self):
        """Determines whether the installer files are stored in a PGA cache

        Returns:
            bool
        """
        if self.url.startswith("N/A"):
            return False
        return has_valid_custom_cache_path()

    @property
    def is_user_pga_caching_allowed(self):
        """Returns true if this file can be transferred to the cache, if
        the user provides it."""
        return self.uses_pga_cache()

    @property
    def _cache_path(self):
        """Return the directory used as a cache for the duration of the installation"""
        return get_url_cache_path(self.url, self.id, self.game_slug)

    def prepare(self):
        """Prepare the file for download. If we've not been redirected to an existing file,
        anwe will create directories to contain the cached file."""
        if not self.is_dest_file_overridden:
            get_url_cache_path(self.url, self.id, self.game_slug, prepare=True)

    def create_download_progress_box(self):
        return DownloadProgressBox(
            url=self.url, dest=self.dest_file, temp=self.download_file, referer=self.referer, downloader=self.downloader
        )

    def check_hash(self):
        """Checks the checksum of `file` and compare it to `value`

        Args:
            checksum (str): The checksum to look for (type:hash)
            dest_file (str): The path to the destination file
            dest_file_uri (str): The uri for the destination file
        """
        if not self.checksum or not self.dest_file:
            return
        try:
            hash_type, expected_hash = self.checksum.split(":", 1)
        except ValueError as err:
            raise ScriptingError(
                _("Invalid checksum, expected format (type:hash) "), faulty_data=self.checksum
            ) from err

        logger.info("Checking hash %s for %s", hash_type, self.dest_file)
        calculated_hash = system.get_file_checksum(self.dest_file, hash_type)
        if calculated_hash != expected_hash:
            raise ScriptingError(
                hash_type.capitalize() + _(" checksum mismatch "), faulty_data=f"{expected_hash} != {calculated_hash}"
            )

    @property
    def size(self) -> Optional[int]:
        if isinstance(self._file_meta, dict) and "size" in self._file_meta:
            try:
                size = int(self._file_meta["size"])
                if size >= 0:
                    return size
            except (ValueError, TypeError):
                return None
        return None

    @property
    def total_size(self) -> Optional[int]:
        if isinstance(self._file_meta, dict) and "total_size" in self._file_meta:
            try:
                total_size = int(self._file_meta["total_size"])
                if total_size >= 0:
                    return total_size
            except (ValueError, TypeError):
                return None
        return None

    def is_ready(self, provider):
        """Is the file already present at the destination (if applicable)?"""
        return provider not in ("user", "pga") or system.path_exists(self.dest_file)

    @property
    def is_cached(self):
        """Is the file available in the local PGA cache?"""
        return self.uses_pga_cache() and system.path_exists(self.dest_file)

    def save_to_cache(self):
        """Copy the file into the PGA cache."""

        cache_path = self._cache_path
        try:
            if not os.path.isdir(cache_path):
                logger.debug("Creating cache path %s", self._cache_path)
                os.makedirs(cache_path)
        except (OSError, PermissionError) as ex:
            logger.error("Failed to created cache path: %s", ex)
            return

        save_to_cache(self.dest_file, cache_path)

    def remove_previous(self):
        """Remove file at already at destination, prior to starting the download."""
        if not self.uses_pga_cache() and system.path_exists(self.dest_file):
            # If we've previously downloaded a directory, we'll need to get rid of it
            # to download a file now. Since we are not using the cache, we don't keep
            # these files anyway - so it should be safe to just nuke and pave all this.
            if os.path.isdir(self.dest_file):
                system.delete_folder(self.dest_file)
            else:
                os.remove(self.dest_file)



"""EA App service."""

import json
import os
import ssl
from gettext import gettext as _
from typing import Any, Dict, Optional
from xml.etree import ElementTree

import requests
import urllib3
from gi.repository import Gio

from lutris import settings
from lutris.config import LutrisConfig, write_game_config
from lutris.database.games import add_game, get_game_by_field
from lutris.database.services import ServiceGameCollection
from lutris.game import Game
from lutris.services.base import SERVICE_LOGIN, OnlineService
from lutris.services.lutris import sync_media
from lutris.services.service_game import ServiceGame
from lutris.services.service_media import ServiceMedia
from lutris.util.log import logger
from lutris.util.strings import slugify

SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION = 1 << 18


class EAAppGames:
    ea_games_location = "Program Files/EA Games"

    def __init__(self, prefix_path):
        self.prefix_path = prefix_path
        self.ea_games_path = os.path.join(self.prefix_path, "drive_c", self.ea_games_location)

    def iter_installed_games(self):
        if not os.path.exists(self.ea_games_path):
            return
        for game_folder in os.listdir(self.ea_games_path):
            yield game_folder

    def get_installed_games_content_ids(self):
        installed_game_ids = []
        for game_folder in self.iter_installed_games():
            installer_data_path = os.path.join(self.ea_games_path, game_folder, "__Installer/installerdata.xml")
            if not os.path.exists(installer_data_path):
                logger.warning("No installerdata.xml for %s", game_folder)
                continue
            tree = ElementTree.parse(installer_data_path)
            nodes = tree.find("contentIDs").findall("contentID")
            if not nodes:
                logger.warning("Content ID not found for %s", game_folder)
                continue
            installed_game_ids.append([node.text for node in nodes])
        return installed_game_ids


class EAAppMedia(ServiceMedia):
    service = "ea_app"
    file_patterns = ["%s.jpg"]
    name = NotImplemented

    @property
    def dest_path(self):
        return os.path.join(settings.CACHE_DIR, self.service, self.name)

    def get_media_url(self, details: Dict[str, Any]) -> Optional[str]:
        image = details["baseItem"][self.name]["largestImage"]
        return image.get("path", None) if image is not None else None


class EAAppKeyArt(EAAppMedia):
    name = "keyArt"
    size = (192, 108)


class EAAppPackArt(EAAppMedia):
    name = "packArt"
    size = (135, 240)


class EAAppPrimaryLogo(EAAppMedia):
    name = "primaryLogo"
    size = (200, 100)


class EAAppGame(ServiceGame):
    service = "ea_app"

    @classmethod
    def new_from_api(cls, game):
        ea_game = EAAppGame()
        ea_game.appid = game["contentId"]
        ea_game.slug = game["gameSlug"]
        ea_game.name = game["baseItem"]["title"]
        ea_game.details = json.dumps(game)
        return ea_game


class LegacyRenegotiationHTTPAdapter(requests.adapters.HTTPAdapter):
    """Allow insecure SSL/TLS protocol renegotiation in an HTTP request.

    By default, OpenSSL v3 expects that servers support RFC 5746. Unfortunately,
    accounts.ea.com does not support this TLS extension (from 2010!), causing
    OpenSSL to refuse to connect. This `requests` HTTP Adapter configures
    OpenSSL to allow "unsafe legacy renegotiation", allowing EA Origin to
    connect. This is only intended as a temporary workaround, and should be
    removed as soon as accounts.ea.com is updated to support RFC 5746.

    Using this adapter will reduce the security of the connection. However, the
    impact should be relatively minimal this is only used to connect to EA
    services. See CVE-2009-3555 for more details.

    See #4235 for more information.
    """

    def init_poolmanager(self, connections, maxsize, block=False, **pool_kwargs):
        """Override the default PoolManager to allow insecure renegotiation."""
        # Based off of the default function from `requests`.
        self._pool_connections = connections
        self._pool_maxsize = maxsize
        self._pool_block = block

        ssl_context = ssl.create_default_context()
        ssl_context.options |= SSL_OP_ALLOW_UNSAFE_LEGACY_RENEGOTIATION

        self.poolmanager = urllib3.PoolManager(
            num_pools=connections,
            maxsize=maxsize,
            block=block,
            strict=True,
            ssl_context=ssl_context,
            **pool_kwargs,
        )


class EAAppService(OnlineService):
    """Service class for EA App"""

    id = "ea_app"
    name = _("EA App")
    icon = "ea_app"
    client_installer = "ea-app"
    login_window_width = 460
    login_window_height = 760
    runner = "wine"
    online = True
    medias = {
        "keyArt": EAAppKeyArt,
        "packArt": EAAppPackArt,
        "primaryLogo": EAAppPrimaryLogo,
    }
    default_format = "keyArt"
    cache_path = os.path.join(settings.CACHE_DIR, "ea_app/cache/")
    cookies_path = os.path.join(settings.CACHE_DIR, "ea_app/cookies")
    token_path = os.path.join(settings.CACHE_DIR, "ea_app/auth_token")
    origin_redirect_uri = "https://www.origin.com/views/login.html"
    login_url = "https://www.ea.com/login"
    redirect_uris = ["https://www.ea.com/"]
    origin_login_url = (
        "https://accounts.ea.com/connect/auth"
        "?response_type=code&client_id=ORIGIN_SPA_ID&display=originXWeb/login"
        "&locale=en_US&release_type=prod"
        "&redirect_uri=%s"
    ) % origin_redirect_uri
    api_url = "https://service-aggregation-layer.juno.ea.com/graphql"
    login_user_agent = settings.DEFAULT_USER_AGENT + " QtWebEngine/5.8.0"

    def __init__(self):
        super().__init__()

        self.session = requests.session()
        self.session.mount("https://", LegacyRenegotiationHTTPAdapter())
        self.access_token = self.load_access_token()

    @property
    def api_headers(self):
        headers = {"User-Agent": self.login_user_agent}
        headers.update(self.get_auth_headers())
        return headers

    def is_connected(self):
        return bool(self.access_token)

    def login_callback(self, url):
        self.fetch_access_token()
        SERVICE_LOGIN.fire(self)

    def fetch_access_token(self):
        token_data = self.get_access_token()
        if not token_data:
            raise RuntimeError("Failed to get access token")
        with open(self.token_path, "w", encoding="utf-8") as token_file:
            token_file.write(json.dumps(token_data, indent=2))
        self.access_token = self.load_access_token()

    def load_access_token(self):
        if not os.path.exists(self.token_path):
            return ""
        with open(self.token_path, encoding="utf-8") as token_file:
            token_data = json.load(token_file)
            return token_data.get("access_token", "")

    def fetch_api(self, query, params: dict = None):
        result = self.session.post(
            self.api_url, headers=self.api_headers, json={"query": query, "variables": params or {}}
        ).json()

        if "errors" in result:
            raise RuntimeError("Errors occurred while running an EA api query.", result["errors"])

        return result

    def get_access_token(self):
        """Request an access token from EA"""
        response = self.session.get(
            "https://accounts.ea.com/connect/auth",
            params={
                "client_id": "ORIGIN_JS_SDK",
                "response_type": "token",
                "redirect_uri": "nucleus:rest",
                "prompt": "none",
            },
            cookies=self.load_cookies(),
        )
        response.raise_for_status()
        token_data = response.json()
        return token_data

    def _request_identity(self):
        response = self.session.get(
            "https://gateway.ea.com/proxy/identity/pids/me",
            cookies=self.load_cookies(),
            headers=self.get_auth_headers(),
        )
        return response.json()

    def get_identity(self):
        """Request the user info"""
        identity_data = self._request_identity()
        if identity_data.get("error") == "invalid_access_token":
            logger.warning("Refreshing EA access token")
            self.fetch_access_token()
            identity_data = self._request_identity()
        elif identity_data.get("error"):
            raise RuntimeError("%s (Error code: %s)" % (identity_data["error"], identity_data["error_number"]))

        if "error" in identity_data:
            raise RuntimeError(identity_data["error"])

        player = self.fetch_api("query{me{player{pd psd displayName}}}")["data"]["me"]["player"]
        user_id = player["pd"]
        persona_id = player["psd"]
        user_name = player["displayName"]
        return str(user_id), str(persona_id), str(user_name)

    def load(self):
        user_id, _persona_id, _user_name = self.get_identity()
        games = self.get_library(user_id)
        logger.info("Retrieved %s games from EA library", len(games))
        ea_games = []
        for game in games:
            ea_game = EAAppGame.new_from_api(game)
            ea_game.save()
            ea_games.append(ea_game)
        return ea_games

    def get_library(self, user_id):
        """Load EA library"""
        chunk_size = 100
        games = []
        entitlements = list(
            filter(
                lambda e: e["product"] is not None and e["product"]["baseItem"]["gameType"] == "BASE_GAME",
                self.get_entitlements(user_id),
            )
        )
        for chunk in [entitlements[i : i + chunk_size] for i in range(0, len(entitlements), chunk_size)]:
            games += self.get_games([e["originOfferId"] for e in chunk])
        return games

    def get_games(self, offer_ids):
        """Load game details from EA"""
        result = self.fetch_api(
            """query getOffers($offerIds: [String!]!) {
                legacyOffers(offerIds: $offerIds, locale: "DEFAULT") {
                    offerId: id
                    contentId
                }
                gameProducts(offerIds: $offerIds, locale: "DEFAULT") {
                    items {
                        id
                        originOfferId
                        gameSlug
                        baseItem {
                            keyArt {
                                largestImage { path }
                            }
                            packArt {
                                largestImage { path }
                            }
                            primaryLogo {
                                largestImage { path }
                            }
                            title
                        }
                    }
                }
            }
            """,
            params={"offerIds": offer_ids},
        )

        games = []
        legacy_offers = result["data"].get("legacyOffers")
        game_products = (result["data"].get("gameProducts") or {}).get("items", [])
        by_offer = {p.get("originOfferId"): p for p in game_products if isinstance(p, dict) and p.get("originOfferId")}
        by_product_id = {p.get("id"): p for p in game_products if isinstance(p, dict) and p.get("id")}

        for legacy_offer in legacy_offers:
            if not isinstance(legacy_offer, dict):
                continue
            offer_id = legacy_offer["offerId"]
            content_id = legacy_offer["contentId"]
            if not offer_id:
                continue
            product = (
                by_offer.get(offer_id)
                or (content_id and by_product_id.get(content_id))
                or by_product_id.get(offer_id)
                or {}
            )
            # Certain games will have identification data but without any product information.
            # Skip those.
            if not product:
                continue
            game = {"contentId": content_id}
            game.update(product)
            games.append(game)
        return games

    def get_entitlements(self, user_id):
        """Request the user's entitlements"""
        games = []
        variables = {"limit": 100}
        while True:
            result = self.fetch_api(
                """query getEntitlements($limit: Int, $next: String) {
                    me {
                        ownedGameProducts(
                            locale: "DEFAULT"
                            entitlementEnabled: true
                            storefronts: [EA]
                            type: [DIGITAL_FULL_GAME, PACKAGED_FULL_GAME]
                            platforms: [PC]
                            paging: {
                                limit: $limit,
                                next: $next
                            }
                        ) {
                            next,
                            items {
                                originOfferId
                                product {
                                    baseItem {
                                        gameType
                                    }
                                }
                            }
                        }
                    }
                }""",
                params=variables,
            )

            products = result["data"]["me"]["ownedGameProducts"]
            variables["next"] = products["next"]
            games += products["items"]
            if products["next"] is None:
                break
        return games

    def get_auth_headers(self):
        """Return headers needed to authenticate HTTP requests"""
        if not self.access_token:
            raise RuntimeError("User not authenticated to EA")
        return {
            "Authorization": "Bearer %s" % self.access_token,
            "AuthToken": self.access_token,
            "X-AuthToken": self.access_token,
        }

    def add_installed_games(self):
        ea_app_game = get_game_by_field("ea-app", "slug")
        if not ea_app_game:
            logger.error("EA App is not installed")
        ea_app_prefix = ea_app_game["directory"].split("drive_c")[0]
        if not os.path.exists(os.path.join(ea_app_prefix, "drive_c")):
            logger.error("Invalid install of EA App at %s", ea_app_prefix)
            return
        ea_app_launcher = EAAppGames(ea_app_prefix)
        installed_slugs = []
        for content_ids in ea_app_launcher.get_installed_games_content_ids():
            slug = self.install_from_ea_app(ea_app_game, content_ids)
            if slug:
                installed_slugs.append(slug)
        sync_media(installed_slugs)
        logger.debug("Installed %s EA games", len(installed_slugs))

    def install_from_ea_app(self, ea_game, content_ids):
        offer_id = content_ids[0]
        logger.debug("Installing EA game %s", offer_id)
        service_game = ServiceGameCollection.get_game("ea_app", offer_id)
        if not service_game:
            logger.error("Aborting install, %s is not present in the game library.", offer_id)
            return
        lutris_game_id = slugify(service_game["name"]) + "-" + self.id
        existing_game = get_game_by_field(lutris_game_id, "installer_slug")
        if existing_game:
            return
        game_config = LutrisConfig(game_config_id=ea_game["configpath"]).game_level
        game_config["game"]["args"] = get_launch_arguments(",".join(content_ids))
        configpath = write_game_config(lutris_game_id, game_config)
        slug = self.get_installed_slug(ea_game)
        add_game(
            name=service_game["name"],
            runner=ea_game["runner"],
            slug=slug,
            directory=ea_game["directory"],
            installed=1,
            installer_slug=lutris_game_id,
            configpath=configpath,
            service=self.id,
            service_id=offer_id,
        )
        return slug

    def generate_installer(self, db_game, ea_db_game):
        ea_game = Game(ea_db_game["id"])
        ea_exe = ea_game.config.game_config["exe"]
        if not os.path.isabs(ea_exe):
            ea_exe = os.path.join(ea_game.config.game_config["prefix"], ea_exe)
        return {
            "name": db_game["name"],
            "version": self.name,
            "slug": slugify(db_game["name"]) + "-" + self.id,
            "game_slug": self.get_installed_slug(db_game),
            "runner": self.get_installed_runner_name(db_game),
            "appid": db_game["appid"],
            "script": {
                "requires": self.client_installer,
                "game": {
                    "args": get_launch_arguments(db_game["appid"]),
                },
                "installer": [
                    {
                        "task": {
                            "name": "wineexec",
                            "executable": ea_exe,
                            "args": get_launch_arguments(db_game["appid"]),
                            "prefix": ea_game.config.game_config["prefix"],
                            "description": ("EA App will now open and prompt you to install %s." % db_game["name"]),
                        }
                    }
                ],
            },
        }

    def get_installed_runner_name(self, db_game):
        return self.runner

    def install(self, db_game):
        ea_app_game = get_game_by_field(self.client_installer, "slug")
        application = Gio.Application.get_default()
        if not ea_app_game or not ea_app_game["installed"]:
            logger.warning("Installing the EA App client")
            application.show_lutris_installer_window(game_slug=self.client_installer)
        else:
            application.show_installer_window(
                [self.generate_installer(db_game, ea_app_game)], service=self, appid=db_game["appid"]
            )


def get_launch_arguments(content_id, action="launch"):
    """Return launch argument for EA games.
    download used to be a valid action but it doesn't seem like it's implemented in EA App."""
    return f"origin2://game/{action}?offerIds={content_id}&autoDownload=1"



"""MAME service
Not ready yet"""

from gettext import gettext as _

from lutris.services.base import BaseService


class MAMEService(BaseService):
    """Service class for MAME"""

    id = "mame"
    name = _("MAME")
    icon = "mame"



"""Utility to get the path of Lutris assets"""

# Standard Library
import os
import sys

# Lutris Modules
from lutris.util import system


def get():
    """Return the path for the resources."""
    launch_path = os.path.realpath(sys.path[0])
    if launch_path.startswith("/usr/local") and os.path.isdir("/usr/local/share/lutris"):
        data_path = "/usr/local/share/lutris"
    elif launch_path.startswith("/usr") and os.path.isdir("/usr/share/lutris"):
        data_path = "/usr/share/lutris"
    elif system.path_exists(os.path.normpath(os.path.join(sys.path[0], "share"))):
        data_path = os.path.normpath(os.path.join(sys.path[0], "share/lutris"))
    elif system.path_exists(os.path.normpath(os.path.join(launch_path, "../../share/lutris"))):
        data_path = os.path.normpath(os.path.join(launch_path, "../../share/lutris"))
    else:
        import lutris

        lutris_module = lutris.__file__
        data_path = os.path.join(os.path.dirname(os.path.dirname(lutris_module)), "share/lutris")
    if not system.path_exists(data_path):
        raise IOError("data_path can't be found at : %s" % data_path)
    return data_path



from typing import Callable, Optional

from gi.repository import Gtk

from lutris import settings
from lutris.gui.config.widget_generator import WidgetWarningMessageBox
from lutris.gui.widgets.common import VBox


class BaseConfigBox(VBox):
    settings_accelerators = {}

    def __init__(self):
        super().__init__(visible=True, spacing=12)
        self.accelerators = None
        self.set_margin_top(50)
        self.set_margin_bottom(50)
        self.set_margin_right(80)
        self.set_margin_left(80)

    def get_section_label(self, text: str) -> Gtk.Label:
        label = Gtk.Label(visible=True)
        label.set_markup("<b>%s</b>" % text)
        label.set_alignment(0, 0.5)
        return label

    def get_description_label(self, text: str) -> Gtk.Label:
        label = Gtk.Label(visible=True)
        label.set_markup("%s" % text)
        label.set_line_wrap(True)
        label.set_alignment(0, 0.5)
        return label

    def _get_framed_options_list_box(self, items):
        frame = Gtk.Frame(visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN)

        list_box = Gtk.ListBox(visible=True, selection_mode=Gtk.SelectionMode.NONE)
        frame.add(list_box)

        for item in items:
            list_box.add(Gtk.ListBoxRow(child=item, visible=True, activatable=False))
        return frame

    def get_setting_box(
        self,
        setting_key: str,
        label: str,
        default: bool = False,
        accelerator: str = None,
        warning_markup: str = None,
        warning_condition: Callable[[bool], bool] = None,
        extra_widget: Gtk.Widget = None,
    ) -> Gtk.Box:
        setting_value = settings.read_bool_setting(setting_key, default=default)

        if not warning_markup and not extra_widget:
            box = self._get_inner_settings_box(setting_key, setting_value, label, accelerator)
        else:
            if warning_markup:

                def update_warning(active):
                    visible = warning_condition(active) if bool(warning_condition) else active
                    warning_box.show_markup(warning_markup if visible else None)

                warning_box = WidgetWarningMessageBox("dialog-warning", margin_left=0, margin_right=0, margin_bottom=0)
                update_warning(setting_value)
                inner_box = self._get_inner_settings_box(
                    setting_key, setting_value, label, accelerator, margin=0, when_setting_changed=update_warning
                )
            else:
                warning_box = None
                inner_box = self._get_inner_settings_box(
                    setting_key,
                    setting_value,
                    label,
                    accelerator,
                    margin=0,
                )

            box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6, visible=True)
            box.pack_start(inner_box, False, False, 0)
            if warning_box:
                box.pack_start(warning_box, False, False, 0)
            if extra_widget:
                box.pack_start(extra_widget, False, False, 0)

        box.set_margin_top(12)
        box.set_margin_bottom(12)
        box.set_margin_left(12)
        box.set_margin_right(12)
        return box

    def _get_inner_settings_box(
        self,
        setting_key: str,
        setting_value: bool,
        label: str,
        accelerator: str = None,
        margin: int = 12,
        when_setting_changed: Callable[[bool], None] = None,
    ):
        checkbox = Gtk.Switch(visible=True, valign=Gtk.Align.CENTER)
        checkbox.set_active(setting_value)
        checkbox.connect("state-set", self.on_setting_change, setting_key, when_setting_changed)

        if accelerator:
            key, mod = Gtk.accelerator_parse(accelerator)
            checkbox.add_accelerator("activate", self.accelerators, key, mod, Gtk.AccelFlags.VISIBLE)

        return self.get_listed_widget_box(label, checkbox, margin=margin)

    def get_listed_widget_box(self, label: str, widget: Gtk.Widget, margin: int = 12) -> Gtk.Box:
        box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12, margin=margin, visible=True)
        label = Gtk.Label(label=label, visible=True, wrap=True)
        label.set_alignment(0, 0.5)
        box.pack_start(label, True, True, 0)
        box.pack_end(widget, False, False, 0)
        return box

    def on_setting_change(
        self, _widget, state: bool, setting_key: str, when_setting_changed: Optional[Callable[[bool], None]] = None
    ) -> None:
        """Save a setting when an option is toggled"""
        settings.write_setting(setting_key, state)
        if when_setting_changed:
            when_setting_changed(state)



# Standard Library
import os
from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class o2em(Runner):
    human_name = _("O2EM")
    description = _("Magnavox Odyssey² Emulator")
    platforms = (
        _("Magnavox Odyssey²"),
        _("Phillips C52"),
        _("Phillips Videopac+"),
        _("Brandt Jopac"),
    )
    bios_path = os.path.expanduser("~/.o2em/bios")
    runner_executable = "o2em/o2em"

    checksums = {
        "o2rom": "562d5ebf9e030a40d6fabfc2f33139fd",
        "c52": "f1071cdb0b6b10dde94d3bc8a6146387",
        "jopac": "279008e4a0db2dc5f1c048853b033828",
        "g7400": "79008e4a0db2dc5f1c048853b033828",
    }

    bios_choices = [
        (_("Magnavox Odyssey²"), "o2rom"),
        (_("Phillips C52"), "c52"),
        (_("Phillips Videopac+"), "g7400"),
        (_("Brandt Jopac"), "jopac"),
    ]
    controller_choices = [
        (_("Disable"), "0"),
        (_("Arrow Keys and Right Shift"), "1"),
        (_("W,S,A,D,SPACE"), "2"),
        (_("Joystick"), "3"),
    ]
    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "default_path": "game_path",
            "help": _("The game data, commonly called a ROM image."),
        }
    ]
    runner_options = [
        {
            "option": "bios",
            "type": "choice",
            "choices": bios_choices,
            "label": _("BIOS"),
            "default": "o2rom",
        },
        {
            "option": "controller1",
            "type": "choice",
            "choices": controller_choices,
            "section": _("Controllers"),
            "label": _("First controller"),
            "default": "2",
        },
        {
            "option": "controller2",
            "type": "choice",
            "choices": controller_choices,
            "section": _("Controllers"),
            "label": _("Second controller"),
            "default": "1",
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "default": False,
        },
        {
            "option": "scanlines",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Scanlines display style"),
            "default": False,
            "help": _("Activates a display filter adding scanlines to imitate the displays of yesteryear."),
        },
    ]

    def get_platform(self):
        bios = self.runner_config.get("bios")
        if bios:
            for i, b in enumerate(self.bios_choices):
                if b[1] == bios:
                    return self.platforms[i]
        return ""

    def install(self, install_ui_delegate, version=None, callback=None):
        def on_runner_installed(*args):
            if not system.path_exists(self.bios_path):
                os.makedirs(self.bios_path)
            if callback:
                callback()

        super().install(install_ui_delegate, version, on_runner_installed)

    def play(self):
        arguments = ["-biosdir=%s" % self.bios_path]

        if self.runner_config.get("fullscreen"):
            arguments.append("-fullscreen")

        if self.runner_config.get("scanlines"):
            arguments.append("-scanlines")

        if "controller1" in self.runner_config:
            arguments.append("-s1=%s" % self.runner_config["controller1"])
        if "controller2" in self.runner_config:
            arguments.append("-s2=%s" % self.runner_config["controller2"])
        rom_path = self.game_config.get("main_file") or ""
        if not system.path_exists(rom_path):
            raise MissingGameExecutableError(filename=rom_path)
        romdir = os.path.dirname(rom_path)
        romfile = os.path.basename(rom_path)
        arguments.append("-romdir=%s/" % romdir)
        arguments.append(romfile)
        return {"command": self.get_command() + arguments}



"""Module for detecting the availability of the Linux futex
FUTEX_WAIT_MULTIPLE operation, or the Linux futex2 syscalls.

Either of these is required for fsync to work in Wine. Fsync is an
alternative implementation of the Windows synchronization primitives
that are used to guard data from being accessed by multiple threads
concurrently (which would be A Bad Thing™).

Fsync improves upon the previous implementation in Wine of these
primitives, known as esync, which in turn improved upon the original
implementation known as "Server-side synchronization".

The original implementation used a wineserver call for each
synchronization operation, which required multiple context switches per
operation.

Esync instead used file descriptors for synchronization, which can be
passed around between processes and therefore allowed synchronization
to happen directly between the processes involved, instead of going
through the wineserver. This made the synchronization operations
faster and improved performance of games a bit.
A problem with this implementation was that each created synchronization
object required one file descriptor, and there is only a limited amount
of these available for each process. Some games would run out of
available file descriptors, and would stop working. This has been partly
mitigated by raising the per-process file descriptor limit, but there
are also games that leak synchronization objects continuously while
running, and would eventually run out despite the raised limits.

Fsync improved on esync by not requiring a file descriptor for each
created synchronization object, and instead using the Linux kernel's
futex interface for synchronizations. This matches Windows's
implementation more closely and mitigated all the file descriptor
related issues of esync. However, since the default futex interface was
insufficient for implementing all required synchronization operations,
a patch to the Linux kernel was needed, which usually meant that users
needed to compile their own Linux kernel with the patch, or install a
kernel provided by a third-party. It was attempted to get the kernel
patch into the mainline Linux kernel, but it didn't get accepted.

Instead, patches were written that would add a new set of system calls
which extend the original futex system calls, dubbed "futex2", and the
Wine fsync code was adjusted to make use of these new system calls.
The new Wine fsync code is backwards-compatible with the old futex
patch, therefore it makes sense for now to detect the presence of
either patch in the running kernel. The detection of the old patch
can probably be removed when the new patch is merged and in a stable
Linux release.

This module's code is based on https://gist.github.com/openglfreak/715d5ab5902497378f1996061dbbf8ec
"""

import ctypes
import errno
import os
import subprocess

__all__ = ("get_fsync_support",)

from lutris.util import cache_single


# pylint: disable=invalid-name,too-few-public-methods
class timespec(ctypes.Structure):
    """Linux kernel compatible timespec type.

    Fields:
        tv_sec: The whole seconds of the timespec.
        tv_nsec: The nanoseconds of the timespec.
    """

    __slots__ = ()
    _fields_ = [
        ("tv_sec", ctypes.c_long),
        ("tv_nsec", ctypes.c_long),
    ]


def _get_syscall_nr_from_headers(syscall_name):
    bits = ctypes.sizeof(ctypes.c_void_p) * 8

    try:
        with subprocess.Popen(
            ("cpp", "-m" + str(bits), "-E", "-P", "-x", "c", "-"),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            universal_newlines=True,
        ) as popen:
            stdout, stderr = popen.communicate("#include <sys/syscall.h>\n__NR_" + syscall_name + "\n")
    except FileNotFoundError as ex:
        raise RuntimeError(
            "failed to determine " + syscall_name + " syscall number: cpp not installed or not in PATH"
        ) from ex

    if popen.returncode:
        raise RuntimeError(
            "failed to determine " + syscall_name + " syscall number: cpp returned nonzero exit code", stderr
        )

    if not stdout:
        raise RuntimeError("failed to determine " + syscall_name + " syscall number: no output from cpp")

    last_line = stdout.splitlines()[-1]

    if last_line == "__NR_futex":
        raise RuntimeError(
            "failed to determine " + syscall_name + " syscall number: __NR_" + syscall_name + " not expanded"
        )

    try:
        return int(last_line)
    except ValueError as ex:
        raise RuntimeError(
            "failed to determine " + syscall_name + " syscall number: "
            "__NR_" + syscall_name + " not a valid number: " + last_line
        ) from ex


# Hardcode some of the most commonly used architectures's
# futex syscall numbers.
_NR_FUTEX_PER_ARCH = {
    ("i386", 32): 240,
    ("i686", 32): 240,
    ("x86_64", 32): 240,
    ("x86_64", 64): 202,
    ("aarch64", 64): 240,
    ("aarch64_be", 64): 240,
    ("armv8b", 32): 240,
    ("armv8l", 32): 240,
}


def _get_futex_syscall_nr():
    """Get the syscall number of the Linux futex() syscall.

    Returns:
        The futex() syscall number.

    Raises:
        RuntimeError: When the syscall number could not be determined.
    """
    bits = ctypes.sizeof(ctypes.c_void_p) * 8

    try:
        return _NR_FUTEX_PER_ARCH[(os.uname()[4], bits)]
    except KeyError:
        pass

    return _get_syscall_nr_from_headers("futex")


def _is_ctypes_obj(obj):
    return hasattr(obj, "_b_base_") and hasattr(obj, "_b_needsfree_") and hasattr(obj, "_objects")


def _is_ctypes_obj_pointer(obj):
    return hasattr(obj, "_type_") and hasattr(obj, "contents")


def _coerce_to_pointer(obj):
    if obj is None:
        return None

    if _is_ctypes_obj(obj):
        if _is_ctypes_obj_pointer(obj):
            return obj
        return ctypes.pointer(obj)

    obj = tuple(obj)
    return (obj[0].__class__ * len(obj))(*obj)


def _get_futex_syscall():
    """Create a function that can be used to execute the Linux futex()
    syscall.

    Returns:
        A proxy function for the Linux futex() syscall.

    Raises:
        AttributeError: When the libc has no syscall() function.
        RuntimeError: When the syscall number could not be determined.
    """
    futex_syscall = ctypes.CDLL(None, use_errno=True).syscall
    futex_syscall.argtypes = (
        ctypes.c_long,
        ctypes.c_void_p,
        ctypes.c_int,
        ctypes.c_int,
        ctypes.POINTER(timespec),
        ctypes.c_void_p,
        ctypes.c_int,
    )
    futex_syscall.restype = ctypes.c_int
    futex_syscall_nr = _get_futex_syscall_nr()

    # pylint: disable=too-many-arguments
    def _futex_syscall(uaddr, futex_op, val, timeout, uaddr2, val3):
        """Invoke the Linux futex() syscall with the provided arguments.

        Args:
            See the description of the futex() syscall for the parameter
            meanings.
            `uaddr` and `uaddr2` are automatically converted to pointers.
            If timeout is None, a zero timeout is passed.

        Returns:
            A tuple of the return value of the syscall and the error code
            in case an error occurred.

        Raises:
            AttributeError: When the libc has no syscall() function.
            RuntimeError: When the syscall number could not be determined.
            TypeError: If `uaddr` or `uaddr2` is not a pointer and can't be
                converted into one.
        """
        error = futex_syscall(
            futex_syscall_nr,
            _coerce_to_pointer(uaddr),
            futex_op,
            val,
            _coerce_to_pointer(timeout or timespec()),
            _coerce_to_pointer(uaddr2),
            val3,
        )
        return error, (ctypes.get_errno() if error == -1 else 0)

    return _futex_syscall


def _get_futex_wait_multiple_op(futex_syscall):
    """Detects which (if any) futex opcode is used for the
    FUTEX_WAIT_MULTIPLE operation on this kernel.

    Returns:
        The opcode number, or None if the operation is not supported.
    """
    ret = futex_syscall(None, 31, 0, None, None, 0)
    if ret[1] != errno.ENOSYS:
        return 31
    ret = futex_syscall(None, 13, 0, None, None, 0)
    if ret[1] != errno.ENOSYS:
        return 13
    return None


@cache_single
def is_futex_wait_multiple_supported():
    """Checks whether the Linux futex FUTEX_WAIT_MULTIPLE operation is
    supported on this kernel.

    Returns:
        Whether this kernel supports the FUTEX_WAIT_MULTIPLE operation.
    """
    try:
        return _get_futex_wait_multiple_op(_get_futex_syscall()) is not None
    except (AttributeError, RuntimeError):
        return False


@cache_single
def is_futex2_supported():
    """Checks whether the Linux futex2 syscall is supported on this
    kernel.

    Returns:
        Whether this kernel supports the futex2 syscall.
    """
    try:
        for filename in ("wait", "waitv", "wake"):
            with open("/sys/kernel/futex2/" + filename, "rb") as file:
                if not file.readline().strip().isdigit():
                    return False
    except OSError:
        return False
    return True


# Hardcode some of the most commonly used architectures's
# futex_waitv syscall numbers.
_NR_FUTEX_WAITV_PER_ARCH = {
    ("i386", 32): 449,
    ("i686", 32): 449,
    ("x86_64", 32): 449,
    ("x86_64", 64): 449,
    ("aarch64", 64): 449,
    ("aarch64_be", 64): 449,
    ("armv8b", 32): 449,
    ("armv8l", 32): 449,
}


def _get_futex_waitv_syscall_nr():
    """Get the syscall number of the Linux futex_waitv() syscall.

    Returns:
        The futex_waitv() syscall number.

    Raises:
        RuntimeError: When the syscall number could not be determined.
    """
    bits = ctypes.sizeof(ctypes.c_void_p) * 8

    try:
        return _NR_FUTEX_WAITV_PER_ARCH[(os.uname()[4], bits)]
    except KeyError:
        pass

    return _get_syscall_nr_from_headers("futex_waitv")


# pylint: disable=invalid-name,too-few-public-methods
class futex_waitv(ctypes.Structure):
    """Linux kernel compatible futex_waitv type.

    Fields:
        val: The expected value.
        uaddr: The address to wait for.
        flags: The type and size of the futex.
    """

    __slots__ = ()
    _fields_ = [
        ("val", ctypes.c_uint64),
        ("uaddr", ctypes.c_void_p),
        ("flags", ctypes.c_uint),
    ]


def _get_futex_waitv_syscall():
    """Create a function that can be used to execute the Linux
    futex_waitv() syscall.

    Returns:
        A proxy function for the Linux futex_waitv() syscall.

    Raises:
        AttributeError: When the libc has no syscall() function.
        RuntimeError: When the syscall number could not be determined.
    """
    futex_waitv_syscall = ctypes.CDLL(None, use_errno=True).syscall
    futex_waitv_syscall.argtypes = (
        ctypes.c_long,
        ctypes.POINTER(futex_waitv),
        ctypes.c_uint,
        ctypes.c_uint,
        ctypes.POINTER(timespec),
    )
    futex_waitv_syscall.restype = ctypes.c_long
    futex_waitv_syscall_nr = _get_futex_waitv_syscall_nr()

    # pylint: disable=too-many-arguments
    def _futex_waitv_syscall(waiters, nr_futexes, flags, timeout):
        """Invoke the Linux futex_waitv() syscall with the provided
        arguments.

        Args:
            See the description of the futex_waitv() syscall for the
            parameter meanings.
            `waiters` is automatically converted to a pointer.
            If timeout is None, a zero timeout is passed.

        Returns:
            A tuple of the return value of the syscall and the error code
            in case an error occurred.

        Raises:
            AttributeError: When the libc has no syscall() function.
            RuntimeError: When the syscall number could not be determined.
            TypeError: If `waiters` is not a pointer and can't be
                converted into one.
        """
        error = futex_waitv_syscall(
            futex_waitv_syscall_nr, _coerce_to_pointer(waiters), nr_futexes, flags, _coerce_to_pointer(timeout)
        )
        return error, (ctypes.get_errno() if error == -1 else 0)

    return _futex_waitv_syscall


@cache_single
def is_futex_waitv_supported():
    """Checks whether the Linux 5.16 futex_waitv syscall is supported on
    this kernel.

    Returns:
        Whether this kernel supports the futex_waitv syscall.
    """
    try:
        ret = _get_futex_waitv_syscall()(None, 0, 0, None)
        return ret[1] != errno.ENOSYS
    except (AttributeError, RuntimeError):
        return False


@cache_single
def get_fsync_support():
    """Checks whether the FUTEX_WAIT_MULTIPLE operation, the futex2
    syscalls, or the futex_waitv syscall is supported on this kernel.

    Returns:
        The result of the check.
    """
    if is_futex_waitv_supported():
        return True
    if is_futex2_supported():
        return True
    if is_futex_wait_multiple_supported():
        return True
    return False



import dataclasses as dc
import json
from typing import List, Optional

import requests


class DataclassJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if dc.is_dataclass(o):
            return dc.asdict(o)
        return super().default(o)


@dc.dataclass
class WebsiteAuthData(object):
    cookie_jar: requests.cookies.RequestsCookieJar
    access_token: str
    region: str


@dc.dataclass(frozen=True)
class BlizzardGame:
    uid: str
    name: str
    family: str


@dc.dataclass(frozen=True)
class ClassicGame(BlizzardGame):
    registry_path: Optional[str] = None
    registry_installation_key: Optional[str] = None
    exe: Optional[str] = None
    bundle_id: Optional[str] = None


@dc.dataclass
class RegionalGameInfo:
    uid: str
    try_for_free: bool


@dc.dataclass
class ConfigGameInfo(object):
    uid: str
    uninstall_tag: Optional[str]
    last_played: Optional[str]


@dc.dataclass
class ProductDbInfo(object):
    uninstall_tag: str
    ngdp: str = ""
    install_path: str = ""
    version: str = ""
    playable: bool = False
    installed: bool = False


class Singleton(type):
    _instances = {}  # type: ignore

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class _Blizzard(object, metaclass=Singleton):
    TITLE_ID_MAP = {
        21297: RegionalGameInfo("s1", True),
        21298: RegionalGameInfo("s2", True),
        5730135: RegionalGameInfo("wow", True),
        5272175: RegionalGameInfo("prometheus", False),
        22323: RegionalGameInfo("w3", False),
        1146311730: RegionalGameInfo("destiny2", False),
        1465140039: RegionalGameInfo("hs_beta", True),
        1214607983: RegionalGameInfo("heroes", True),
        17459: RegionalGameInfo("diablo3", True),
        4613486: RegionalGameInfo("fenris", True),
        1447645266: RegionalGameInfo("viper", False),
        1329875278: RegionalGameInfo("odin", True),
        1279351378: RegionalGameInfo("lazarus", False),
        1514493267: RegionalGameInfo("zeus", False),
        1381257807: RegionalGameInfo("rtro", False),
        1464615513: RegionalGameInfo("wlby", False),
        5198665: RegionalGameInfo("osi", False),
        1179603525: RegionalGameInfo("fore", False),
    }
    TITLE_ID_MAP_CN = {**TITLE_ID_MAP, 17459: RegionalGameInfo("d3cn", False)}
    BATTLENET_GAMES = [
        BlizzardGame("s1", "StarCraft", "S1"),
        BlizzardGame("s2", "StarCraft II", "S2"),
        BlizzardGame("wow", "World of Warcraft", "WoW"),
        BlizzardGame("wow_classic", "World of Warcraft Classic", "WoW_wow_classic"),
        BlizzardGame("prometheus", "Overwatch", "Pro"),
        BlizzardGame("w3", "Warcraft III", "W3"),
        BlizzardGame("hs_beta", "Hearthstone", "WTCG"),
        BlizzardGame("heroes", "Heroes of the Storm", "Hero"),
        BlizzardGame("d3cn", "暗黑破壞神III", "D3CN"),
        BlizzardGame("diablo3", "Diablo III", "D3"),
        BlizzardGame("fenris", "Diablo IV", "FEN"),
        BlizzardGame("viper", "Call of Duty: Black Ops 4", "VIPR"),
        BlizzardGame("odin", "Call of Duty: Modern Warfare", "ODIN"),
        BlizzardGame("lazarus", "Call of Duty: MW2 Campaign Remastered", "LAZR"),
        BlizzardGame("zeus", "Call of Duty: Black Ops Cold War", "ZEUS"),
        BlizzardGame("rtro", "Blizzard Arcade Collection", "RTRO"),
        BlizzardGame("wlby", "Crash Bandicoot 4: It's About Time", "WLBY"),
        BlizzardGame("osi", "Diablo® II: Resurrected", "OSI"),
        BlizzardGame("fore", "Call of Duty: Vanguard", "FORE"),
    ]
    CLASSIC_GAMES = [
        ClassicGame("d2", "Diablo® II", "Diablo II", "Diablo II", "DisplayIcon", "Game.exe", "com.blizzard.diabloii"),
        ClassicGame("d2LOD", "Diablo® II: Lord of Destruction®", "Diablo II"),  # TODO exe and bundleid
        ClassicGame(
            "w3ROC",
            "Warcraft® III: Reign of Chaos",
            "Warcraft III",
            "Warcraft III",
            "InstallLocation",
            "Warcraft III.exe",
            "com.blizzard.WarcraftIII",
        ),
        ClassicGame(
            "w3tft",
            "Warcraft® III: The Frozen Throne®",
            "Warcraft III",
            "Warcraft III",
            "InstallLocation",
            "Warcraft III.exe",
            "com.blizzard.WarcraftIII",
        ),
        ClassicGame("sca", "StarCraft® Anthology", "Starcraft", "StarCraft"),  # TODO exe and bundleid
    ]

    def __init__(self):
        self._games = {game.uid: game for game in self.BATTLENET_GAMES + self.CLASSIC_GAMES}

    def __getitem__(self, key: str) -> BlizzardGame:
        """
        :param key: str uid (eg. "prometheus")
        :returns: game by `key`
        """
        return self._games[key]

    def game_by_title_id(self, title_id: int, cn: bool) -> BlizzardGame:
        """
        :param cn: flag if china game definitions should be search though
        :raises KeyError: when unknown title_id for given region
        """
        if cn:
            regional_info = self.TITLE_ID_MAP_CN[title_id]
        else:
            regional_info = self.TITLE_ID_MAP[title_id]
        return self[regional_info.uid]

    def try_for_free_games(self, cn: bool) -> List[BlizzardGame]:
        """
        :param cn: flag if china game definitions should be search though
        """
        return [
            self[info.uid] for info in (self.TITLE_ID_MAP_CN if cn else self.TITLE_ID_MAP).values() if info.try_for_free
        ]


Blizzard = _Blizzard()



from lutris import settings
from lutris.api import get_api_games
from lutris.database import sql
from lutris.database.games import get_games
from lutris.util.log import logger


def migrate():
    """Add blank sortname field to games that do not yet have one"""
    logger.info("Adding blank sortname field to database")
    slugs_to_update = [game["slug"] for game in get_games()]
    games = get_api_games(slugs_to_update)
    for game in games:
        if "sortname" not in game.keys() or game["sortname"] is None:
            sql.db_update(settings.DB_PATH, "games", {"sortname": ""}, {"slug": game["slug"]})
        logger.info("Added blank sortname for %s", game["name"])



"""Various utilities using the GObject framework"""

import array
import os
from typing import TYPE_CHECKING, Iterable, List, Optional, Type, TypeVar, Union, cast

import cairo
from gi.repository import Gdk, GdkPixbuf, Gio, GLib, Gtk

from lutris import settings
from lutris.exceptions import MissingMediaError
from lutris.gui.widgets import NotificationSource
from lutris.util import datapath, magic, system
from lutris.util.log import logger

if TYPE_CHECKING:
    from lutris.gui.application import LutrisApplication
    from lutris.gui.lutriswindow import LutrisWindow

try:
    from PIL import Image
except ImportError:
    Image = None

ICON_SIZE = (32, 32)
BANNER_SIZE = (184, 69)
MEDIA_CACHE_INVALIDATED = NotificationSource()


def get_application() -> Optional["LutrisApplication"]:
    return cast("LutrisApplication", Gio.Application.get_default())


def get_required_application() -> "LutrisApplication":
    application = cast("LutrisApplication", Gio.Application.get_default())
    if not application:
        raise RuntimeError("The LutrisApplication does not exist.")
    return application


def get_main_window() -> Optional["LutrisWindow"]:
    """Return the application's main window, or None if it doesn't exist
    (though it almost alway does exist)"""
    application = get_application()
    return application.window if application else None


def get_required_main_window() -> "LutrisWindow":
    """Return the application's main window or raises an exception if
    it doesn't exist."""
    application = get_required_application()
    window = application.window
    if not window:
        raise RuntimeError("The main window does not exist.")
    return window


def get_widget_window(widget: Optional[Gtk.Widget]) -> Optional[Gtk.Window]:
    """Returns the window that contains a widget, if any. This wll return None
    for a widget that is not in a window, rather than returning the widget itself
    like get_toplevel()."""
    if widget:
        return cast(Optional[Gtk.Window], widget.get_ancestor(Gtk.Window))
    else:
        return None


TChildWidget = TypeVar("TChildWidget", bound=Gtk.Widget)


def get_widget_children(
    widget: Optional[Gtk.Widget], child_type: Optional[Type[TChildWidget]] = None
) -> List[TChildWidget]:
    """Returns the children of any widget; non-containers have no children
    and returns an empty list. This can filter out a specific type of child widget if child_type
    is not None, but otherwise it returns all children."""
    if isinstance(widget, Gtk.Container):
        if child_type:
            return [cast(TChildWidget, w) for w in widget.get_children() if isinstance(w, child_type)]
        else:
            return list(cast(Iterable[TChildWidget], widget.get_children()))
    else:
        return []


def open_uri(uri):
    """Opens a local or remote URI with the default application"""
    system.spawn(["xdg-open", uri])


def get_image_file_extension(path: str) -> Optional[str]:
    """Returns the canonical file extension for an image,
    either 'jpg' or 'png'; we deduce this from the file extension, or if that fails the
    file's 'magic' prefix bytes."""
    ext = os.path.splitext(path)[1].casefold()
    if ext in [".jpg", ".jpeg"]:
        return ".jpg"
    if ext == ".png":
        return ".png"

    try:
        file_type = magic.from_file(path).casefold()
    except OSError:
        return None  # file is missing, or can't read it

    if "jpeg image data" in file_type:
        return ".jpg"
    if "png image data" in file_type:
        return ".png"

    return None


def get_surface_size(surface):
    """Returns the size of a surface, accounting for the device scale;
    the surface's get_width() and get_height() are in physical pixels."""
    device_scale_x, device_scale_y = surface.get_device_scale()
    width = surface.get_width() / device_scale_x
    height = surface.get_height() / device_scale_y
    return width, height


def get_scaled_surface_by_path(path, size, device_scale, preserve_aspect_ratio=True):
    """Returns a Cairo surface containing the image at the path given. It has the size indicated.

    You specify the device_scale, and the bitmap is generated at an enlarged size accordingly,
    but with the device scale of the surface also set; in this way a high-DPI image can be
    rendered conveniently.

    If you pass True for preserve_aspect_ratio, the aspect ratio of the image is preserved,
    but will be no larger than the size (times the device_scale).

    If there's no file at the path, or it is empty, this function returns None.
    """
    pixbuf = get_pixbuf_by_path(path)
    if not pixbuf:
        return None

    pixbuf_width = pixbuf.get_width()
    pixbuf_height = pixbuf.get_height()

    scale_x = (size[0] / pixbuf_width) * device_scale
    scale_y = (size[1] / pixbuf_height) * device_scale

    if preserve_aspect_ratio:
        scale_x = min(scale_x, scale_y)
        scale_y = scale_x

    pixel_width = round(pixbuf_width * scale_x)
    pixel_height = round(pixbuf_height * scale_y)

    surface = cairo.ImageSurface(cairo.Format.ARGB32, pixel_width, pixel_height)  # pylint:disable=no-member
    cr = cairo.Context(surface)  # pylint:disable=no-member
    cr.scale(scale_x, scale_y)
    Gdk.cairo_set_source_pixbuf(cr, pixbuf, 0, 0)
    cr.get_source().set_extend(cairo.Extend.PAD)  # pylint: disable=no-member
    cr.paint()
    surface.set_device_scale(device_scale, device_scale)
    return surface


def get_default_icon_path(size):
    """Returns the path to the default icon for the size given; it's
    a Lutris icon for a square size, and a gradient for other sizes."""
    if not size or size[0] == size[1]:
        filename = "media/default_icon.png"
    else:
        filename = "media/default_banner.png"
    return os.path.join(datapath.get(), filename)


def get_pixbuf_by_path(path, size=None, preserve_aspect_ratio=True):
    """Reads an image file and returns the pixbuf. If you provide a size, this scales
    the file to fit that size, preserving the aspect ratio if preserve_aspect_ratio is
    True. If the file is missing or empty, or if 'path' is None or empty,
    this returns None. Still raises GLib.GError for corrupt files."""
    if not system.path_exists(path, exclude_empty=True):
        return None

    if size:
        # new_from_file_at_size scales but preserves aspect ratio
        width, height = size
        if preserve_aspect_ratio:
            return GdkPixbuf.Pixbuf.new_from_file_at_size(path, width, height)

        return GdkPixbuf.Pixbuf.new_from_file_at_scale(path, width, height, preserve_aspect_ratio=False)

    return GdkPixbuf.Pixbuf.new_from_file(path)


def get_required_pixbuf_by_path(path, size=None, preserve_aspect_ratio=True):
    """Reads an image file and returns the pixbuf. If you provide a size, this scales
    the file to fit that size, preserving the aspect ratio if preserve_aspect_ratio is
    True. If the file is missing or unreadable, or if 'path' is None or empty, this raises
    MissingMediaError."""
    try:
        pixbuf = get_pixbuf_by_path(path, size, preserve_aspect_ratio)
        if not pixbuf:
            raise MissingMediaError(filename=path)
        return pixbuf
    except GLib.GError as ex:
        logger.exception("Unable to load icon from image %s", path)
        raise MissingMediaError(message=str(ex), filename=path) from ex


def has_stock_icon(name):
    """This tests if a GTK stock icon is known; if not we can try a fallback."""
    if not name:
        return False

    theme = Gtk.IconTheme.get_default()
    return theme.has_icon(name)


def pick_stock_icon(names: Union[str, Iterable[str]], fallback_name: str = "package-x-generic-symbolic") -> str:
    """Used to select a stock icon that actually exists in the icon set; it tries all the names given,
    and returns the first where has_stock_icon is true. If none pass the test, returns fallback_name instead."""
    if isinstance(names, str):
        names = [names]

    for name in names:
        if has_stock_icon(name):
            return name

    return fallback_name


def get_runtime_icon_path(icon_name):
    """Finds the icon file for an icon whose name is given; this searches the icons
    in Lutris's runtime directory. The name is normalized by removing spaces
    and lower-casing it, and both .png and .svg files with the name can be found.

    Arguments:
    icon_name -- The name of the icon to retrieve

    Returns:
        The path to the icon, or None if it wasn't found.
    """
    filename = icon_name.lower().replace(" ", "")
    # We prefer bitmaps over SVG, because we've got some SVG icons with the
    # wrong size (oops) and this avoids them.
    search_directories = [
        "icons/hicolor/64x64/apps",
        "icons/hicolor/24x24/apps",
        "icons",
        "icons/hicolor/scalable/apps",
        "icons/hicolor/symbolic/apps",
    ]
    extensions = [".png", ".svg"]
    for search_dir in search_directories:
        for ext in extensions:
            icon_path = os.path.join(settings.RUNTIME_DIR, search_dir, filename + ext)
            if os.path.exists(icon_path):
                return icon_path
    return None


def convert_to_background(background_path, target_size=(320, 1080)):
    """Converts an image to a pane background"""
    coverart = Image.open(background_path)
    coverart = coverart.convert("RGBA")

    target_width, target_height = target_size
    image_height = int(target_height * 0.80)  # 80% of the mask is visible
    orig_width, orig_height = coverart.size

    # Resize and crop coverart
    width = int(orig_width * (image_height / orig_height))
    offset = int((width - target_width) / 2)
    coverart = coverart.resize((width, image_height), resample=Image.Resampling.BICUBIC)
    coverart = coverart.crop((offset, 0, target_width + offset, image_height))

    # Resize canvas of coverart by putting transparent pixels on the bottom
    coverart_bg = Image.new("RGBA", (target_width, target_height), (0, 0, 0, 0))
    coverart_bg.paste(coverart, (0, 0, target_width, image_height))

    # Apply a tint to the base image
    # tint = Image.new('RGBA', (target_width, target_height), (0, 0, 0, 255))
    # coverart = Image.blend(coverart, tint, 0.6)

    # Paste coverart on transparent image while applying a gradient mask
    background = Image.new("RGBA", (target_width, target_height), (0, 0, 0, 0))
    mask = Image.open(os.path.join(datapath.get(), "media/mask.png"))
    background.paste(coverart_bg, mask=mask)

    return background


def thumbnail_image(base_image, target_size):
    base_width, base_height = base_image.size
    base_ratio = base_width / base_height
    target_width, target_height = target_size
    target_ratio = target_width / target_height

    # Resize and crop coverart
    if base_ratio >= target_ratio:
        width = int(base_width * (target_height / base_height))
        height = target_height
    else:
        width = target_width
        height = int(base_height * (target_width / base_width))
    x_offset = int((width - target_width) / 2)
    y_offset = int((height - target_height) / 2)
    base_image = base_image.resize((width, height), resample=Image.Resampling.BICUBIC)
    base_image = base_image.crop((x_offset, y_offset, width - x_offset, height - y_offset))
    return base_image


def paste_overlay(base_image, overlay_image, position=0.7):
    base_width, base_height = base_image.size
    overlay_width, overlay_height = overlay_image.size
    offset_x = int((base_width - overlay_width) / 2)
    offset_y = int((base_height - overlay_height) / 2)
    base_image.paste(
        overlay_image, (offset_x, offset_y, overlay_width + offset_x, overlay_height + offset_y), mask=overlay_image
    )
    return base_image


def image2pixbuf(image):
    """Converts a PIL Image to a GDK Pixbuf"""
    image_array = array.array("B", image.tobytes())
    width, height = image.size
    return GdkPixbuf.Pixbuf.new_from_data(image_array, GdkPixbuf.Colorspace.RGB, True, 8, width, height, width * 4)


def load_icon_theme():
    """Add the lutris icon folder to the default theme"""
    icon_theme = Gtk.IconTheme.get_default()
    local_theme_path = os.path.join(settings.RUNTIME_DIR, "icons")
    if local_theme_path not in icon_theme.get_search_path():
        icon_theme.prepend_search_path(local_theme_path)



"""Add, remove and configure runners"""

from gettext import gettext as _

from gi.repository import Gtk

from lutris import runners, settings
from lutris.gui.config.base_config_box import BaseConfigBox
from lutris.gui.config.runner_box import RunnerBox
from lutris.gui.widgets.utils import open_uri
from lutris.search import RunnerSearch


class RunnersBox(BaseConfigBox):
    """List of all available runners"""

    def __init__(self):
        super().__init__()
        self._runner_search = RunnerSearch("")
        self.search_entry_placeholder_text = ""

        self.add(self.get_section_label(_("Add, remove or configure runners")))
        self.add(
            self.get_description_label(
                _("Runners are programs such as emulators, engines or translation layers capable of running games.")
            )
        )
        self.search_failed_label = Gtk.Label(_("No runners matched the search"))
        self.pack_start(self.search_failed_label, False, False, 0)
        self.runner_list_frame = Gtk.Frame(visible=True, shadow_type=Gtk.ShadowType.ETCHED_IN)
        self.runner_listbox = Gtk.ListBox(visible=True)
        self.runner_list_frame.add(self.runner_listbox)
        self.pack_start(self.runner_list_frame, False, False, 0)

    def populate_runners(self):
        runner_count = 0
        for runner_name in sorted(runners.__all__):
            list_box_row = Gtk.ListBoxRow(visible=True)
            list_box_row.set_selectable(False)
            list_box_row.set_activatable(False)
            list_box_row.add(RunnerBox(runner_name))
            self.runner_listbox.add(list_box_row)
            runner_count += 1

        self._update_row_visibility()
        # pretty sure there will always be many runners, so assume plural
        self.search_entry_placeholder_text = _("Search %s runners") % runner_count

    @staticmethod
    def on_folder_clicked(_widget):
        open_uri("file://" + settings.RUNNER_DIR)

    @property
    def filter(self):
        return self._runner_search.text

    @filter.setter
    def filter(self, value):
        self._runner_search = RunnerSearch(value)
        self._update_row_visibility()

    def _update_row_visibility(self):
        search = self._runner_search

        any_matches = False
        for row in self.runner_listbox.get_children():
            runner_box = row.get_child()
            runner = runner_box.runner
            match = search.matches(runner)
            row.set_visible(match)
            if match:
                any_matches = True

        self.runner_list_frame.set_visible(any_matches)
        self.search_failed_label.set_visible(not any_matches)



DEFAULT_INSTALLERS = {
    "3do": {
        "runner": "libretro",
        "game": {"core": "opera", "main_file": "rom"},
    },
    "sms": {
        "runner": "libretro",
        "game": {"core": "genesis_plus_gx", "main_file": "rom"},
    },
    "gg": {
        "runner": "libretro",
        "game": {"core": "genesis_plus_gx", "main_file": "rom"},
    },
    "md": {
        "runner": "libretro",
        "game": {"core": "genesis_plus_gx", "main_file": "rom"},
    },
    "pico": {
        "runner": "libretro",
        "game": {"core": "picodrive", "main_file": "rom"},
    },
    "segacd": {
        "runner": "libretro",
        "game": {"core": "picodrive", "main_file": "rom"},
    },
    "saturn": {
        "runner": "libretro",
        "game": {"core": "yabause", "main_file": "rom"},
    },
    "dc": {
        "runner": "libretro",
        "game": {"core": "flycast", "main_file": "rom"},
    },
    "colecovision": {"runner": "mame", "game": {"main_file": "rom", "machine": "coleco", "device": "cart"}},
    "atari800": {"runner": "atari800", "game": {"main_file": "rom", "machine": "xl"}},
    "atari2600": {"runner": "stella", "game": {"main_file": "rom"}},
    "lynx": {
        "runner": "libretro",
        "game": {"core": "handy", "main_file": "rom"},
    },
    "atari-st": {
        "runner": "hatari",
        "game": {
            "disk-a": "rom",
        },
    },
    "jaguar": {
        "runner": "libretro",
        "game": {"core": "virtualjaguar", "main_file": "rom"},
    },
    "amiga": {"runner": "fsuae", "game": {"main_file": "rom"}},
    "amiga-1200": {"runner": "fsuae", "game": {"main_file": "rom"}, "fsuae": {"model": "A1200"}},
    "amiga-cd32": {"runner": "fsuae", "game": {"main_file": "rom"}, "fsuae": {"model": "CD32"}},
    "ds": {
        "runner": "libretro",
        "game": {"core": "desmume", "main_file": "rom"},
    },
    "gb": {
        "runner": "libretro",
        "game": {"core": "gambatte", "main_file": "rom"},
    },
    "gba": {
        "runner": "libretro",
        "game": {"core": "vba_next", "main_file": "rom"},
    },
    "nes": {
        "runner": "libretro",
        "game": {"core": "mesen", "main_file": "rom"},
    },
    "snes": {
        "runner": "libretro",
        "game": {"core": "snes9x", "main_file": "rom"},
    },
    "n64": {
        "runner": "libretro",
        "game": {"core": "mupen64plus_next", "main_file": "rom"},
    },
    "gamecube": {"runner": "dolphin", "game": {"main_file": "rom", "platform": "0"}},
    "wii": {"runner": "dolphin", "game": {"main_file": "rom", "platform": "1"}},
    "switch": {"runner": "yuzu", "game": {"main_file": "rom"}},
    "ps1": {
        "runner": "libretro",
        "game": {"core": "mednafen_psx_hw", "main_file": "rom"},
    },
    "ps2": {"runner": "pcsx2", "game": {"main_file": "rom"}},
    "ps3": {"runner": "rpcs3", "game": {"main_file": "rom"}},
    "psp": {
        "runner": "libretro",
        "game": {"core": "ppsspp", "main_file": "rom"},
    },
    "cdi": {"runner": "mame", "game": {"main_file": "rom", "device": "cdrm", "machine": "cdimono1"}},
    "msx": {
        "runner": "libretro",
        "game": {"core": "bluemsx", "main_file": "rom"},
    },
    "archimedes": {"runner": "mame", "game": {"main_file": "rom", "device": "flop1", "machine": "aa4000"}},
    "bbc": {"runner": "mame", "game": {"main_file": "rom", "device": "flop1", "machine": "bbcbp128"}},
    "electron": {"runner": "mame", "game": {"main_file": "rom", "device": "flop1", "machine": "electron"}},
    "astrocade": {"runner": "mame", "game": {"main_file": "rom", "device": "cart", "machine": "astrocde"}},
    "wonderswancolor": {"runner": "mame", "game": {"main_file": "rom", "device": "cart", "machine": "wscolor"}},
    "wonderswan": {"runner": "mame", "game": {"main_file": "rom", "device": "cart", "machine": "wswan"}},
    "cpc6128disk": {"runner": "mame", "game": {"main_file": "rom", "device": "flop1", "machine": "cpc6128p"}},
    "gx4000": {"runner": "mame", "game": {"main_file": "rom", "device": "cart", "machine": "gx4000"}},
    "apple2": {"runner": "mame", "game": {"main_file": "rom", "device": "flop1", "machine": "apple2gs"}},
    "spectrumcass": {"runner": "mame", "game": {"main_file": "rom", "device": "cass", "machine": "spectrum"}},
    "c64prg": {"runner": "libretro", "game": {"core": "vice_x64", "main_file": "rom"}},
    "c64g64": {"runner": "libretro", "game": {"core": "vice_x64", "main_file": "rom"}},
}



from gi.repository import Gio, GLib, GObject, Gtk

from lutris import settings
from lutris.gui.widgets import NotificationSource
from lutris.util.log import logger

PORTAL_BUS_NAME = "org.freedesktop.portal.Desktop"
PORTAL_OBJECT_PATH = "/org/freedesktop/portal/desktop"
PORTAL_SETTINGS_INTERFACE = "org.freedesktop.portal.Settings"

THEME_CHANGED = NotificationSource()


class StyleManager(GObject.Object):
    """Manages the color scheme of the app.

    Has a single readable GObject property `is_dark` telling whether the app is
    in dark mode, it is set to True, when either the user preference on the
    preferences panel or in the system is set to prefer dark mode.
    """

    _dbus_proxy = None
    _preferred_theme = "default"
    _system_theme = None
    _is_dark = False

    def __init__(self):
        super().__init__()

        self.gtksettings = Gtk.Settings.get_default()
        self.preferred_theme = settings.read_setting("preferred_theme") or "default"

        Gio.DBusProxy.new_for_bus(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            PORTAL_BUS_NAME,
            PORTAL_OBJECT_PATH,
            PORTAL_SETTINGS_INTERFACE,
            None,
            self._new_for_bus_cb,
        )

    def _read_portal_setting(self) -> None:
        if not self._dbus_proxy:
            return
        variant = GLib.Variant.new_tuple(
            GLib.Variant.new_string("org.freedesktop.appearance"),
            GLib.Variant.new_string("color-scheme"),
        )
        self._dbus_proxy.call(
            "Read",
            variant,
            Gio.DBusCallFlags.NONE,
            GObject.G_MAXINT,
            None,
            self._call_cb,
        )

    def _new_for_bus_cb(self, obj, result):
        try:
            proxy = obj.new_for_bus_finish(result)
            if proxy:
                proxy.connect("g-signal", self._on_settings_changed)
                self._dbus_proxy = proxy
                self._read_portal_setting()
            else:
                raise RuntimeError("Could not start GDBusProxy")
        except Exception as ex:
            logger.exception("Error setting up style change monitoring: %s", ex)

    def _call_cb(self, obj, result):
        try:
            values = obj.call_finish(result)
            if values:
                value = values[0]
                self.system_theme = self._read_value(value)
            else:
                raise RuntimeError("Could not read color-scheme")
        except Exception as ex:
            logger.exception("Error reading color-scheme: %s", ex)

    def _on_settings_changed(self, _proxy, _sender_name, signal_name, params):
        if signal_name != "SettingChanged":
            return

        namespace, name, value = params

        if namespace == "org.freedesktop.appearance" and name == "color-scheme":
            self.system_theme = self._read_value(value)

    def _read_value(self, value: int) -> str:
        if value == 1:
            return "dark"

        if value == 2:
            return "light"

        return "default"

    @property
    def is_config_dark(self) -> bool:
        """True if we override light mode to be dark; if we're
        defaulting to dark, this does nothing."""
        return self._is_config_dark

    @is_config_dark.setter  # type: ignore
    def is_config_dark(self, is_config_dark: bool) -> None:
        if self._is_config_dark == is_config_dark:
            return

        self._is_config_dark = is_config_dark
        self._update_is_dark()

    @property
    def preferred_theme(self) -> str:
        """Can be 'light' or 'dark' to override the theme, or 'default' to go with
        the system's default theme."""
        return self._preferred_theme

    @preferred_theme.setter  # type: ignore
    def preferred_theme(self, preferred_theme: str) -> None:
        if self._preferred_theme == preferred_theme:
            return

        self._preferred_theme = preferred_theme
        self._update_is_dark()

    @property
    def system_theme(self) -> str:
        return self._system_theme or "default"

    @system_theme.setter  # type: ignore
    def system_theme(self, system_theme: str) -> None:
        if self._system_theme == system_theme:
            return

        self._system_theme = system_theme
        self._update_is_dark()

    @GObject.Property(type=bool, default=False, flags=GObject.ParamFlags.READABLE)
    def is_dark(self) -> bool:
        return self._is_dark

    def _update_is_dark(self) -> None:
        if self.is_dark_by_default:
            is_dark = self.preferred_theme != "light"
        else:
            is_dark = self.preferred_theme == "dark"

        if self._is_dark == is_dark:
            return

        self._is_dark = is_dark
        self.notify("is-dark")

        self.gtksettings.set_property("gtk-application-prefer-dark-theme", is_dark)
        THEME_CHANGED.fire()

    @property
    def is_dark_by_default(self):
        return self.system_theme != "light"



"""DuckStation Runner"""

from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system
from lutris.util.log import logger


class duckstation(Runner):
    human_name = _("DuckStation")
    description = _("PlayStation 1 Emulator")
    platforms = [_("Sony PlayStation")]
    runnable_alone = True
    runner_executable = "duckstation/DuckStation-x64.AppImage"
    flatpak_id = "org.duckstation.DuckStation"
    download_url = "https://github.com/stenzek/duckstation/releases/download/latest/DuckStation-x64.AppImage"

    game_options = [
        {
            "option": "main_file",
            "type": "file",
            "label": _("ROM file"),
            "default_path": "game_path",
        }
    ]

    # Duckstation uses an AppImage, no need for the runtime.
    system_options_override = [{"option": "disable_runtime", "default": True}]

    def play(self):
        arguments = self.get_command()

        rom = self.game_config.get("main_file") or ""
        if not system.path_exists(rom):
            raise MissingGameExecutableError(filename=rom)
        arguments.append(rom)
        logger.debug("DuckStation starting with args: %s", arguments)
        return {"command": arguments}



from gi.repository import Gtk

from lutris.gui.widgets.utils import (
    ICON_SIZE,
    get_default_icon_path,
    get_pixbuf_by_path,
    get_required_pixbuf_by_path,
    get_runtime_icon_path,
    has_stock_icon,
)


class ScaledImage(Gtk.Image):
    """This class provides a rather basic feature the GtkImage doesn't offer - the ability
    to scale the image rendered. Scaling a pixbuf is not the same thing - that discards
    pixel data. This will preserve it on high-DPI displays by scaling only at drawing time."""

    __gtype_name__ = "ScaledImage"

    def __init__(self, scale_factor, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.scale_factor = scale_factor

    @staticmethod
    def new_scaled_from_path(path, size=None, preserve_aspect_ratio=True, scale_factor=1):
        """Constructs an image showing the image at the path, scaled to the size given.
        The scale factor is used to scale up the pixbuf, but scale down the image so
        a higher-res image can be shown in the same space on a High-DPI screen. You
        pass your widget's get_scale_factor() here."""

        pixbuf_size = (size[0] * scale_factor, size[1] * scale_factor) if size else None
        pixbuf = get_required_pixbuf_by_path(path, pixbuf_size)
        image = ScaledImage(1 / scale_factor)
        image.set_from_pixbuf(pixbuf)
        return image

    @staticmethod
    def new_from_media_path(path, size, scale_factor=1):
        """Constructs an image showing Lutris media, read from the path, scaled to the
        size given, as with new_scaled_from_path().

        However, if the path is not readable, this will substitute a default icon
        or banner. If 'size' is square, you get the icon; if not it is a gradient
        filling the full size given."""

        pixbuf_size = (size[0] * scale_factor, size[1] * scale_factor)

        pixbuf = get_pixbuf_by_path(path, pixbuf_size)
        if not pixbuf:
            default_icon = get_default_icon_path(size)
            pixbuf = get_required_pixbuf_by_path(default_icon, pixbuf_size, preserve_aspect_ratio=False)

        image = ScaledImage(1 / scale_factor)
        image.set_from_pixbuf(pixbuf)
        return image

    @staticmethod
    def get_runtime_icon_image(icon_name, fallback_stock_icon_name=None, scale_factor=1, visible=False):
        """Returns a ScaledImage of an icon for runtime or service; the image has the
        default icon size. If the icon can't be found, we'll fall back onto another,
        stock icon. If you don't supply one (or it's not available) we'll fall back
        further to 'package-x-generic-symbolic'; we always give you something."""

        path = get_runtime_icon_path(icon_name)
        if path:
            icon = ScaledImage.new_scaled_from_path(path, size=ICON_SIZE, scale_factor=scale_factor)
        else:
            if not has_stock_icon(fallback_stock_icon_name):
                fallback_stock_icon_name = "package-x-generic-symbolic"

            icon = Gtk.Image.new_from_icon_name(fallback_stock_icon_name, Gtk.IconSize.DND)
        icon.set_visible(visible)
        return icon

    def do_get_preferred_width(self):
        minimum, natural = Gtk.Image.do_get_preferred_width(self)
        return minimum * self.scale_factor, natural * self.scale_factor

    def do_get_preferred_height(self):
        minimum, natural = Gtk.Image.do_get_preferred_height(self)
        return minimum * self.scale_factor, natural * self.scale_factor

    def do_draw(self, cr):
        if self.scale_factor != 1:
            # we need to scale around the center of the image,
            # but cr.scale() scales around (0, 0). So we move
            # the co-ordinates before scaling.
            allocation = self.get_allocation()
            center_x = allocation.width / 2
            center_y = allocation.height / 2

            cr.translate(center_x, center_y)
            cr.scale(self.scale_factor, self.scale_factor)
            cr.translate(-center_x, -center_y)

        Gtk.Image.do_draw(self, cr)



# Standard Library
from gettext import gettext as _
from os import path

from lutris.exceptions import DirectoryNotFoundError, GameConfigError, MissingGameExecutableError

# Lutris Modules
from lutris.runners.runner import Runner


class easyrpg(Runner):
    human_name = _("EasyRPG Player")
    description = _("Runs RPG Maker 2000/2003 games")
    platforms = [_("Linux")]
    runnable_alone = True
    entry_point_option = "project_path"
    runner_executable = "easyrpg/easyrpg-player"
    download_url = "https://easyrpg.org/downloads/player/0.8.1.1/easyrpg-player-0.8.1.1-linux.tar.gz"

    game_options = [
        {
            "option": "project_path",
            "type": "directory",
            "label": _("Game directory"),
            "help": _("Select the directory of the game. <b>(required)</b>"),
        },
        {
            "option": "encoding",
            "type": "choice",
            "advanced": True,
            "label": _("Encoding"),
            "help": _(
                "Instead of auto detecting the encoding or using the one in RPG_RT.ini, the specified encoding is used."
            ),
            "choices": [
                (_("Auto"), ""),
                (_("Auto (ignore RPG_RT.ini)"), "auto"),
                (_("Western European"), "1252"),
                (_("Central/Eastern European"), "1250"),
                (_("Japanese"), "932"),
                (_("Cyrillic"), "1251"),
                (_("Korean"), "949"),
                (_("Chinese (Simplified)"), "936"),
                (_("Chinese (Traditional)"), "950"),
                (_("Greek"), "1253"),
                (_("Turkish"), "1254"),
                (_("Hebrew"), "1255"),
                (_("Arabic"), "1256"),
                (_("Baltic"), "1257"),
                (_("Thai"), "874"),
            ],
            "default": "",
        },
        {
            "option": "engine",
            "type": "choice",
            "advanced": True,
            "label": _("Engine"),
            "help": _("Disable auto detection of the simulated engine."),
            "choices": [
                (_("Auto"), ""),
                (_("RPG Maker 2000 engine (v1.00 - v1.10)"), "rpg2k"),
                (_("RPG Maker 2000 engine (v1.50 - v1.51)"), "rpg2kv150"),
                (_("RPG Maker 2000 (English release) engine"), "rpg2ke"),
                (_("RPG Maker 2003 engine (v1.00 - v1.04)"), "rpg2k3"),
                (_("RPG Maker 2003 engine (v1.05 - v1.09a)"), "rpg2k3v105"),
                (_("RPG Maker 2003 (English release) engine"), "rpg2k3e"),
            ],
            "default": "",
        },
        {
            "option": "patch",
            "type": "string",
            "advanced": True,
            "label": _("Patches"),
            "help": _(
                "Instead of autodetecting patches used by this game, force emulation of certain patches.\n"
                "\nAvailable patches:\n"
                '<b>common-this</b>: "This Event" in common events'
                "<b>dynrpg</b>: DynRPG patch by Cherry"
                "<b>key-patch</b>: Key Patch by Ineluki"
                "<b>maniac</b>: Maniac Patch by BingShan"
                "<b>pic-unlock</b>: Pictures are not blocked by messages"
                "<b>rpg2k3-cmds</b>: Support all RPG Maker 2003 event commands in any version of the engine"
                "\n\nYou can provide multiple patches or use 'none' to disable all engine patches."
            ),
        },
        {
            "option": "language",
            "type": "string",
            "advanced": True,
            "label": _("Language"),
            "help": _("Load the game translation in the language/LANG directory."),
        },
        {
            "option": "save_path",
            "type": "directory",
            "label": _("Save path"),
            "warn_if_non_writable_parent": True,
            "help": _(
                "Instead of storing save files in the game directory they are stored in the specified path. "
                "The directory must exist."
            ),
        },
        {
            "option": "new_game",
            "type": "bool",
            "label": _("New game"),
            "help": _("Skip the title scene and start a new game directly."),
            "default": False,
        },
        {
            "option": "load_game_id",
            "type": "range",
            "label": _("Load game ID"),
            "help": _("Skip the title scene and load SaveXX.lsd.\nSet to 0 to disable."),
            "min": 0,
            "max": 99,
            "default": 0,
        },
        {
            "option": "record_input",
            "type": "file",
            "advanced": True,
            "label": _("Record input"),
            "help": _("Records all button input to the specified log file."),
        },
        {
            "option": "replay_input",
            "type": "file",
            "advanced": True,
            "label": _("Replay input"),
            "help": _(
                "Replays button input from the specified log file, as generated by 'Record input'.\n"
                "If the RNG seed and the state of the save file directory is also the same as it was "
                "when the log was recorded, this should reproduce an identical run to the one recorded."
            ),
        },
        {
            "option": "test_play",
            "type": "bool",
            "advanced": True,
            "section": _("Debug"),
            "label": _("Test play"),
            "help": _("Enable TestPlay (debug) mode."),
            "default": False,
        },
        {
            "option": "hide_title",
            "type": "bool",
            "advanced": True,
            "section": _("Debug"),
            "label": _("Hide title"),
            "help": _("Hide the title background image and center the command menu."),
            "default": False,
        },
        {
            "option": "start_map_id",
            "type": "range",
            "advanced": True,
            "section": _("Debug"),
            "label": _("Start map ID"),
            "help": _(
                "Overwrite the map used for new games and use MapXXXX.lmu instead.\n"
                "Set to 0 to disable.\n\n"
                "Incompatible with 'Load game ID'."
            ),
            "min": 0,
            "max": 9999,
            "default": 0,
        },
        {
            "option": "start_position",
            "type": "string",
            "advanced": True,
            "section": _("Debug"),
            "label": _("Start position"),
            "help": _(
                "Overwrite the party start position and move the party to the specified position.\n"
                "Provide two numbers separated by a space.\n\n"
                "Incompatible with 'Load game ID'."
            ),
        },
        {
            "option": "start_party",
            "type": "string",
            "advanced": True,
            "section": _("Debug"),
            "label": _("Start party"),
            "help": _(
                "Overwrite the starting party members with the actors with the specified IDs.\n"
                "Provide one to four numbers separated by spaces.\n\n"
                "Incompatible with 'Load game ID'."
            ),
        },
        {
            "option": "battle_test",
            "type": "string",
            "advanced": True,
            "section": _("Debug"),
            "label": _("Battle test"),
            "help": _("Start a battle test with the specified monster party."),
        },
    ]

    runner_options = [
        {
            "option": "autobattle_algo",
            "type": "choice",
            "advanced": True,
            "section": _("Engine"),
            "label": _("AutoBattle algorithm"),
            "help": _(
                "Which AutoBattle algorithm to use.\n\n"
                "<b>RPG_RT</b>: The default RPG_RT compatible algorithm, including RPG_RT bugs.\n"
                "<b>RPG_RT+</b>: The default RPG_RT compatible algorithm, with bug-fixes.\n"
                "<b>ATTACK</b>: Like RPG_RT+ but only physical attacks, no skills."
            ),
            "choices": [
                (_("Auto"), ""),
                (_("RPG_RT"), "RPG_RT"),
                (_("RPG_RT+"), "RPG_RT+"),
                (_("ATTACK"), "ATTACK"),
            ],
            "default": "",
        },
        {
            "option": "enemyai_algo",
            "type": "choice",
            "advanced": True,
            "section": _("Engine"),
            "label": _("EnemyAI algorithm"),
            "help": _(
                "Which EnemyAI algorithm to use.\n\n"
                "<b>RPG_RT</b>: The default RPG_RT compatible algorithm, including RPG_RT bugs.\n"
                "<b>RPG_RT+</b>: The default RPG_RT compatible algorithm, with bug-fixes.\n"
            ),
            "choices": [
                (_("Auto"), ""),
                (_("RPG_RT"), "RPG_RT"),
                (_("RPG_RT+"), "RPG_RT+"),
            ],
            "default": "",
        },
        {
            "option": "seed",
            "type": "range",
            "advanced": True,
            "section": _("Engine"),
            "label": _("RNG seed"),
            "help": _("Seeds the random number generator.\nUse -1 to disable."),
            "min": -1,
            "max": 2147483647,
            "default": -1,
        },
        {
            "option": "audio",
            "type": "bool",
            "section": _("Audio"),
            "label": _("Enable audio"),
            "help": _("Switch off to disable audio."),
            "default": True,
        },
        {
            "option": "music_volume",
            "type": "range",
            "section": _("Audio"),
            "label": _("BGM volume"),
            "help": _("Volume of the background music."),
            "min": 0,
            "max": 100,
            "default": 100,
        },
        {
            "option": "sound_volume",
            "type": "range",
            "section": _("Audio"),
            "label": _("SFX volume"),
            "help": _("Volume of the sound effects."),
            "min": 0,
            "max": 100,
            "default": 100,
        },
        {
            "option": "soundfont",
            "type": "file",
            "advanced": True,
            "section": _("Audio"),
            "label": _("Soundfont"),
            "help": _("Soundfont in sf2 format to use when playing MIDI files."),
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Fullscreen"),
            "help": _("Start in fullscreen mode."),
            "default": False,
        },
        {
            "option": "game_resolution",
            "type": "choice",
            "section": _("Graphics"),
            "advanced": True,
            "label": _("Game resolution"),
            "help": _(
                "Force a different game resolution.\n\nThis is experimental and can cause glitches or break games!"
            ),
            "choices": [
                (_("320×240 (4:3, Original)"), "original"),
                (_("416×240 (16:9, Widescreen)"), "widescreen"),
                (_("560×240 (21:9, Ultrawide)"), "ultrawide"),
            ],
            "default": "original",
        },
        {
            "option": "scaling",
            "type": "choice",
            "section": _("Graphics"),
            "label": _("Scaling"),
            "help": _(
                "How the video output is scaled.\n\n"
                "<b>Nearest</b>: Scale to screen size (causes scaling artifacts)\n"
                "<b>Integer</b>: Scale to multiple of the game resolution\n"
                "<b>Bilinear</b>: Like Nearest, but output is blurred to avoid artifacts\n"
            ),
            "choices": [
                (_("Nearest"), "nearest"),
                (_("Integer"), "integer"),
                (_("Bilinear"), "bilinear"),
            ],
            "default": "bilinear",
        },
        {
            "option": "stretch",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Stretch"),
            "help": _("Ignore the aspect ratio and stretch video output to the entire width of the screen."),
            "default": False,
        },
        {
            "option": "vsync",
            "type": "bool",
            "section": _("Graphics"),
            "label": _("Enable VSync"),
            "help": _("Switch off to disable VSync and use the FPS limit."),
            "default": True,
        },
        {
            "option": "fps_limit",
            "type": "range",
            "section": _("Graphics"),
            "label": _("FPS limit"),
            "help": _(
                "Set a custom frames per second limit.\n"
                "If unspecified, the default is 60 FPS.\n"
                "Set to 0 to disable the frame limiter."
            ),
            "min": 0,
            "max": 9999,
            "default": 60,
        },
        {
            "option": "show_fps",
            "type": "choice",
            "section": _("Graphics"),
            "label": _("Show FPS"),
            "help": _("Enable frames per second counter."),
            "choices": [
                (_("Disabled"), "off"),
                (_("Fullscreen & title bar"), "on"),
                (_("Fullscreen, title bar & window"), "full"),
            ],
            "default": "off",
        },
        {
            "option": "rtp",
            "type": "bool",
            "section": _("Runtime Package"),
            "label": _("Enable RTP"),
            "help": _("Switch off to disable support for the Runtime Package (RTP)."),
            "default": True,
        },
        {
            "option": "rpg2k_rtp_path",
            "type": "directory",
            "section": _("Runtime Package"),
            "label": _("RPG2000 RTP location"),
            "help": _("Full path to a directory containing an extracted RPG Maker 2000 Run-Time-Package (RTP)."),
        },
        {
            "option": "rpg2k3_rtp_path",
            "type": "directory",
            "section": _("Runtime Package"),
            "label": _("RPG2003 RTP location"),
            "help": _("Full path to a directory containing an extracted RPG Maker 2003 Run-Time-Package (RTP)."),
        },
        {
            "option": "rpg_rtp_path",
            "type": "directory",
            "section": _("Runtime Package"),
            "label": _("Fallback RTP location"),
            "help": _("Full path to a directory containing a combined RTP."),
        },
    ]

    @property
    def game_path(self):
        game_path = self.game_data.get("directory")
        if game_path:
            return path.expanduser(game_path)  # just in case

        # Default to the directory of the entry point
        entry_point = self.game_config.get(self.entry_point_option)
        if entry_point:
            return path.expanduser(entry_point)

        return ""

    def get_env(self, os_env=False, disable_runtime=False):
        env = super().get_env(os_env, disable_runtime=disable_runtime)

        rpg2k_rtp_path = self.runner_config.get("rpg2k_rtp_path")
        if rpg2k_rtp_path:
            env["RPG2K_RTP_PATH"] = rpg2k_rtp_path

        rpg2k3_rtp_path = self.runner_config.get("rpg2k3_rtp_path")
        if rpg2k3_rtp_path:
            env["RPG2K3_RTP_PATH"] = rpg2k3_rtp_path

        rpg_rtp_path = self.runner_config.get("rpg_rtp_path")
        if rpg_rtp_path:
            env["RPG_RTP_PATH"] = rpg_rtp_path

        return env

    def get_command(self):
        cmd = super().get_command()

        # Engine
        autobattle_algo = self.runner_config.get("autobattle_algo")
        if autobattle_algo:
            cmd.extend(("--autobattle-algo", autobattle_algo))

        enemyai_algo = self.runner_config.get("enemyai_algo")
        if enemyai_algo:
            cmd.extend(("--enemyai-algo", enemyai_algo))

        seed = self.runner_config.get("seed")
        if seed:
            cmd.extend(("--seed", str(seed)))

        # Audio
        if not self.runner_config["audio"]:
            cmd.append("--no-audio")

        music_volume = self.runner_config.get("music_volume")
        if music_volume:
            cmd.extend(("--music-volume", str(music_volume)))

        sound_volume = self.runner_config.get("sound_volume")
        if sound_volume:
            cmd.extend(("--sound-volume", str(sound_volume)))

        soundfont = self.runner_config.get("soundfont")
        if soundfont:
            cmd.extend(("--soundfont", soundfont))

        # Graphics
        if self.runner_config["fullscreen"]:
            cmd.append("--fullscreen")
        else:
            cmd.append("--window")

        game_resolution = self.runner_config.get("game_resolution")
        if game_resolution:
            cmd.extend(("--game-resolution", game_resolution))

        scaling = self.runner_config.get("scaling")
        if scaling:
            cmd.extend(("--scaling", scaling))

        if self.runner_config["stretch"]:
            cmd.append("--stretch")

        if not self.runner_config["vsync"]:
            cmd.append("--no-vsync")

        fps_limit = self.runner_config.get("fps_limit")
        if fps_limit:
            cmd.extend(("--fps-limit", str(fps_limit)))

        show_fps = self.runner_config.get("show_fps")
        if show_fps != "off":
            cmd.append("--show-fps")
        if show_fps == "full":
            cmd.append("--fps-render-window")

        # Runtime Package
        if not self.runner_config["rtp"]:
            cmd.append("--no-rtp")

        return cmd

    def get_run_data(self):
        cmd = self.get_command()

        if self.default_path:
            game_path = path.expanduser(self.default_path)
            cmd.extend(("--project-path", game_path))

        return {"command": cmd, "env": self.get_env()}

    def play(self):
        if not self.game_path:
            raise GameConfigError(_("No game directory provided"))
        if not path.isdir(self.game_path):
            raise DirectoryNotFoundError(directory=self.game_path)

        cmd = self.get_command()

        cmd.extend(("--project-path", self.game_path))

        encoding = self.game_config.get("encoding")
        if encoding:
            cmd.extend(("--encoding", encoding))

        engine = self.game_config.get("engine")
        if engine:
            cmd.extend(("--engine", engine))

        patches = self.game_config.get("patches")
        if patches == "none":
            cmd.append("--no-patch")
        elif patches:
            cmd.extend(("--patches", *patches.split()))

        language = self.game_config.get("language")
        if language:
            cmd.extend(("--language", language))

        save_path = self.game_config.get("save_path")
        if save_path:
            save_path = path.expanduser(save_path)
            if not path.isdir(save_path):
                raise DirectoryNotFoundError(directory=self.game_path)
            cmd.extend(("--save-path", save_path))

        record_input = self.game_config.get("record_input")
        if record_input:
            record_input = path.expanduser(record_input)
            cmd.extend(("--record-input", record_input))

        replay_input = self.game_config.get("replay_input")
        if replay_input:
            replay_input = path.expanduser(replay_input)
            if not path.isfile(replay_input):
                raise MissingGameExecutableError(filename=replay_input)
            cmd.extend(("--replay-input", replay_input))

        load_game_id = self.game_config.get("load_game_id")
        if load_game_id:
            cmd.extend(("--load-game-id", str(load_game_id)))

        # Debug
        if self.game_config["test_play"]:
            cmd.append("--test-play")

        if self.game_config["hide_title"]:
            cmd.append("--hide-title")

        start_map_id = self.game_config.get("start_map_id")
        if start_map_id:
            cmd.extend(("--start-map-id", str(start_map_id)))

        start_position = self.game_config.get("start_position")
        if start_position:
            cmd.extend(("--start-position", *start_position.split()))

        start_party = self.game_config.get("start_party")
        if start_party:
            cmd.extend(("--start-party", *start_party.split()))

        battle_test = self.game_config.get("battle_test")
        if battle_test:
            cmd.extend(("--battle-test", battle_test))

        return {"command": cmd}



# Third Party Libraries
from gi.repository import Gtk


class LogTextView(Gtk.TextView):
    # pylint: disable=no-member

    def __init__(self, buffer=None, autoscroll=True, wrap_mode=Gtk.WrapMode.CHAR):
        super().__init__(visible=True)

        if buffer:
            self.set_buffer(buffer)
        self.set_editable(False)
        self.set_cursor_visible(False)
        self.set_monospace(True)
        self.set_left_margin(10)
        self.scroll_max = 0
        self.set_wrap_mode(wrap_mode)
        self.get_style_context().add_class("lutris-logview")

        self.mark = self.create_new_mark(self.props.buffer.get_start_iter())

        if autoscroll:
            self.connect("size-allocate", self.autoscroll)

    def autoscroll(self, *args):  # pylint: disable=unused-argument
        adj = self.get_vadjustment()
        if adj.get_value() == self.scroll_max or self.scroll_max == 0:
            adj.set_value(adj.get_upper() - adj.get_page_size())
            self.scroll_max = adj.get_value()
        else:
            self.scroll_max = adj.get_upper() - adj.get_page_size()

    def create_new_mark(self, buffer_iter):
        return self.props.buffer.create_mark(None, buffer_iter, True)

    def reset_search(self):
        self.props.buffer.delete_mark(self.mark)
        self.mark = self.create_new_mark(self.props.buffer.get_start_iter())
        self.props.buffer.place_cursor(self.props.buffer.get_iter_at_mark(self.mark))

    def find_first(self, searched_entry):
        self.reset_search()
        self.find_next(searched_entry)

    def find_next(self, searched_entry):
        buffer_iter = self.props.buffer.get_iter_at_mark(self.mark)
        next_occurence = buffer_iter.forward_search(
            searched_entry.get_text(), Gtk.TextSearchFlags.CASE_INSENSITIVE, None
        )

        # Found nothing try from the beginning
        if next_occurence is None:
            next_occurence = self.props.buffer.get_start_iter().forward_search(
                searched_entry.get_text(), Gtk.TextSearchFlags.CASE_INSENSITIVE, None
            )

        # Highlight if result
        if next_occurence is not None:
            self.highlight(next_occurence[0], next_occurence[1])
            self.props.buffer.delete_mark(self.mark)
            self.mark = self.create_new_mark(next_occurence[1])

    def find_previous(self, searched_entry):
        # First go to the beginning of searched_entry string
        buffer_iter = self.props.buffer.get_iter_at_mark(self.mark)
        buffer_iter.backward_chars(len(searched_entry.get_text()))

        previous_occurence = buffer_iter.backward_search(
            searched_entry.get_text(), Gtk.TextSearchFlags.CASE_INSENSITIVE, None
        )

        # Found nothing ? Try from the end
        if previous_occurence is None:
            previous_occurence = self.props.buffer.get_end_iter().backward_search(
                searched_entry.get_text(), Gtk.TextSearchFlags.CASE_INSENSITIVE, None
            )

        # Highlight if result
        if previous_occurence is not None:
            self.highlight(previous_occurence[0], previous_occurence[1])
            self.props.buffer.delete_mark(self.mark)
            self.mark = self.create_new_mark(previous_occurence[1])

    def highlight(self, range_start, range_end):
        self.props.buffer.select_range(range_start, range_end)
        # Focus
        self.scroll_mark_onscreen(self.mark)



# Standard Library
from gettext import gettext as _

from lutris.exceptions import DirectoryNotFoundError

# Lutris Modules
from lutris.runners.runner import Runner
from lutris.util import system


class cemu(Runner):
    human_name = _("Cemu")
    platforms = [_("Wii U")]
    description = _("Wii U emulator")
    runnable_alone = True
    runner_executable = "cemu/Cemu"
    flatpak_id = "info.cemu.Cemu"
    game_options = [
        {
            "option": "main_file",
            "type": "directory",
            "label": _("Game directory"),
            "help": _(
                "The directory in which the game lives. "
                "If installed into Cemu, this will be in the mlc directory, such as mlc/usr/title/00050000/101c9500."
            ),
        },
        {
            "option": "wua_rom",
            "type": "file",
            "label": _("Compressed ROM"),
            "help": _("A game compressed into a single file (WUA format), only use if not using game directory"),
        },
    ]
    runner_options = [
        {
            "option": "fullscreen",
            "label": _("Fullscreen"),
            "type": "bool",
            "default": True,
        },
        {
            "option": "mlc",
            "label": _("Custom mlc folder location"),
            "type": "directory",
        },
        {
            "option": "ud",
            "label": _("Render in upside down mode"),
            "type": "bool",
            "default": False,
            "advanced": True,
        },
        {
            "option": "nsight",
            "label": _("NSight debugging options"),
            "type": "bool",
            "default": False,
            "advanced": True,
        },
        {
            "option": "legacy",
            "label": _("Intel legacy graphics mode"),
            "type": "bool",
            "default": False,
            "advanced": True,
        },
    ]

    def play(self):
        """Run the game."""
        arguments = self.get_command()

        fullscreen = self.runner_config.get("fullscreen")
        if fullscreen:
            arguments.append("-f")
        mlc = self.runner_config.get("mlc")
        if mlc:
            if not system.path_exists(mlc):
                raise DirectoryNotFoundError(directory=mlc)
            arguments += ["-m", mlc]
        ud = self.runner_config.get("ud")
        if ud:
            arguments.append("-u")
        nsight = self.runner_config.get("nsight")
        if nsight:
            arguments.append("--nsight")
        legacy = self.runner_config.get("legacy")
        if legacy:
            arguments.append("--legacy")
        gamedir = self.game_config.get("main_file") or self.game_config.get("wua_rom") or ""
        if not system.path_exists(gamedir):
            raise DirectoryNotFoundError(directory=gamedir)
        arguments += ["-g", gamedir]
        return {"command": arguments}



"""ShadPS4 Runner"""

import os
from gettext import gettext as _

from lutris.exceptions import MissingGameExecutableError
from lutris.runners.runner import Runner
from lutris.util import system


class shadps4(Runner):
    human_name = _("ShadPS4")
    description = _("PlayStation 4 Emulator")
    platforms = [_("Sony PlayStation 4")]
    runnable_alone = True
    runner_executable = "shadps4/shadPS4QtLauncher-qt.AppImage"

    game_options = [
        {
            "option": "main_file",
            "type": "directory",
            "label": _("Game folder"),
            "help": _("Path to the game folder (e.g., CUSA00001)"),
            "default_path": "game_path",
        }
    ]

    runner_options = [
        {
            "option": "emulator",
            "type": "choice",
            "label": _("Emulator"),
            "choices": [
                (_("Bundled"), "bundled"),
                (_("Launcher default"), "default"),
                (_("Custom"), "custom"),
            ],
            "default": "bundled",
        },
        {
            "option": "custom_emulator_path",
            "type": "file",
            "label": _("Custom emulator path"),
            "help": _("Path to ShadPS4 SDL emulator (only used if 'Custom' is selected)"),
            "advanced": True,
        },
        {
            "option": "fullscreen",
            "type": "bool",
            "label": _("Fullscreen"),
            "default": True,
        },
    ]

    system_options_override = [{"option": "disable_runtime", "default": True}]

    def get_sdl_emulator_path(self):
        """Derive the SDL emulator path from the launcher path."""
        launcher = self.get_executable()
        launcher_dir = os.path.dirname(launcher)
        # Look for the SDL AppImage in the same directory
        for name in ("Shadps4-sdl.AppImage", "shadps4-sdl.AppImage", "shadPS4-sdl.AppImage"):
            sdl_path = os.path.join(launcher_dir, name)
            if system.path_exists(sdl_path):
                return sdl_path
        return None

    def play(self):
        arguments = self.get_command()

        # Emulator selection
        emu_choice = self.runner_config.get("emulator", "bundled")
        if emu_choice == "bundled":
            emu_path = self.get_sdl_emulator_path()
            if emu_path:
                arguments.extend(["-e", emu_path])
            else:
                arguments.append("-d")  # Fall back to launcher default
        elif emu_choice == "custom":
            custom_path = self.runner_config.get("custom_emulator_path")
            if custom_path:
                arguments.extend(["-e", custom_path])
            else:
                arguments.append("-d")
        else:  # default
            arguments.append("-d")

        # Game
        game_path = self.game_config.get("main_file") or ""
        if not system.path_exists(game_path):
            raise MissingGameExecutableError(filename=game_path)
        arguments.extend(["-g", game_path])

        # Emulator args after separator
        arguments.append("--")
        arguments.extend(["--fullscreen", "true" if self.runner_config.get("fullscreen") else "false"])

        return {"command": arguments}



from lutris.util.wine.dll_manager import DLLManager


class dgvoodoo2Manager(DLLManager):
    name = "dgvoodoo2"
    human_name = "dgvoodoo2"
    managed_dlls = (
        "d3dimm",
        "ddraw",
        "glide",
        "glide2x",
        "glide3x",
    )
    managed_appdata_files = ["dgVoodoo/dgVoodoo.conf"]
    releases_url = "https://api.github.com/repos/lutris/dgvoodoo2/releases"
    proton_compatible = True



"""Linux specific platform code"""

import json
import os
import platform
import re
import resource
import shutil
import sys
from collections import Counter, defaultdict
from gettext import gettext as _
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union, cast

from lutris import settings
from lutris.exceptions import MisconfigurationError
from lutris.util import cache_single, flatpak, system
from lutris.util.graphics import drivers, glxinfo, vkquery
from lutris.util.log import logger

try:
    import distro
except ImportError:
    logger.warning("Package 'distro' unavailable. Unable to read Linux distribution")
    distro = None

# Linux components used by lutris
SYSTEM_COMPONENTS = {
    "COMMANDS": [
        "xrandr",
        "fuser",
        "glxinfo",
        "vulkaninfo",
        "fuser",
        "7z",
        "gtk-update-icon-cache",
        "lspci",
        "ldconfig",
    ],
    "OPTIONAL_COMMANDS": [
        "fluidsynth",
        "nvidia-smi",
        "fluidsynth",
    ],
    "TERMINALS": [
        "xterm",
        "gnome-terminal",
        "konsole",
        "xfce4-terminal",
        "pantheon-terminal",
        "terminator",
        "mate-terminal",
        "urxvt",
        "cool-retro-term",
        "Eterm",
        "guake",
        "lilyterm",
        "lxterminal",
        "roxterm",
        "rxvt",
        "aterm",
        "sakura",
        "st",
        "terminology",
        "termite",
        "tilix",
        "wterm",
        "kitty",
        "yuakuake",
        "qterminal",
        "alacritty",
        "kgx",
        "deepin-terminal",
        "wezterm",
        "foot",
        "ptyxis",
    ],
    "LIBRARIES": {
        "OPENGL": ["libGL.so.1"],
        "VULKAN": ["libvulkan.so.1"],
        "WINE": ["libsqlite3.so.0"],
        "RADEON": ["libvulkan_radeon.so"],
        "GAMEMODE": ["libgamemodeauto.so"],
        "GNUTLS": ["libgnutls.so.30"],
    },
}


@cache_single
def is_exherbo_with_cross_i686() -> bool:
    return system.path_exists("/etc/exherbo-release") and system.path_exists("/etc/ld-i686-pc-linux-gnu.cache")


class LinuxSystem:  # pylint: disable=too-many-public-methods
    """Global cache for system commands"""

    _cache = {}

    multiarch_lib_folders = [
        ("/lib", "/lib64"),
        ("/lib32", "/lib64"),
        ("/usr/lib", "/usr/lib64"),
        ("/usr/lib32", "/usr/lib64"),
        ("/lib/i386-linux-gnu", "/lib/x86_64-linux-gnu"),
        ("/usr/lib/i386-linux-gnu", "/usr/lib/x86_64-linux-gnu"),
        ("/usr/lib", "/opt/32/lib"),
        ("/usr/i686-pc-linux-gnu/lib", "/usr/x86_64-pc-linux-gnu/lib"),
    ]

    soundfont_folders = [
        "/usr/share/sounds/sf2",
        "/usr/share/soundfonts",
    ]

    recommended_no_file_open = 524288
    required_components = ["OPENGL", "VULKAN", "GNUTLS"]
    optional_components = ["WINE", "GAMEMODE"]

    def __init__(self) -> None:
        for key in ("COMMANDS", "OPTIONAL_COMMANDS", "TERMINALS"):
            self._cache[key] = {}
            for command in SYSTEM_COMPONENTS[key]:
                command_path = shutil.which(command)
                if not command_path:
                    command_path = self.get_sbin_path(command)
                if command_path:
                    self._cache[key][command] = command_path
                elif key == "COMMANDS":
                    logger.warning("Command '%s' not found on your system", command)

        # Detect if system is 64bit capable
        self.is_64_bit = sys.maxsize > 2**32
        self.arch = self.get_arch()
        self.shared_libraries = self.get_shared_libraries()
        self.populate_libraries()
        self.populate_sound_fonts()
        self.soft_limit, self.hard_limit = self.get_file_limits()
        self.glxinfo = self.get_glxinfo()

    @staticmethod
    def get_sbin_path(command: str) -> Optional[str]:
        """Some distributions don't put sbin directories in $PATH"""
        path_candidates = ["/sbin", "/usr/sbin"]
        for candidate in path_candidates:
            command_path = os.path.join(candidate, command)
            if os.path.exists(command_path):
                return command_path
        return None

    @staticmethod
    def get_file_limits() -> Tuple[int, int]:
        return resource.getrlimit(resource.RLIMIT_NOFILE)

    def has_enough_file_descriptors(self) -> bool:
        return self.hard_limit >= self.recommended_no_file_open

    @staticmethod
    def get_cpus() -> List[Dict[str, str]]:
        """Parse the output of /proc/cpuinfo"""
        cpus = [{}]
        cpu_index = 0
        with open("/proc/cpuinfo", encoding="utf-8") as cpuinfo:
            for line in cpuinfo.readlines():
                if not line.strip():
                    cpu_index += 1
                    cpus.append({})
                    continue
                key, value = line.split(":", 1)
                cpus[cpu_index][key.strip()] = value.strip()
        return [cpu for cpu in cpus if cpu]

    @staticmethod
    def get_drives() -> List[Dict[str, Any]]:
        """Return a list of drives with their filesystems"""
        lsblk_output = system.read_process_output(["lsblk", "-f", "--json"])
        if not lsblk_output:
            return []
        return [drive for drive in json.loads(lsblk_output)["blockdevices"] if drive["fstype"] != "squashfs"]

    @staticmethod
    def get_ram_info() -> Dict[str, str]:
        """Parse the output of /proc/meminfo and return RAM information in kB"""
        mem = {}
        with open("/proc/meminfo", encoding="utf-8") as meminfo:
            for line in meminfo.readlines():
                key, value = line.split(":", 1)
                mem[key.strip()] = value.strip("kB \n")
        return mem

    def get_dist_info(self) -> Union[str, Tuple[str, str, str]]:
        """Return distribution information"""
        if distro is None:
            return "unknown"
        # In Flatpak, read host's os-release instead of the runtime's
        if self.is_flatpak():
            host_distro = distro.LinuxDistribution(root_dir="/run/host")
            return host_distro.name(), host_distro.version(), host_distro.codename()
        return distro.linux_distribution()

    @staticmethod
    def get_arch() -> Optional[str]:
        """Return the system architecture only if compatible
        with the supported architectures from the Lutris API
        """
        machine = platform.machine()
        if machine == "x86_64":
            return "x86_64"
        if machine in ("i386", "i686"):
            return "i386"
        if "armv7" in machine:
            return "armv7"
        logger.warning("Unsupported architecture %s", machine)
        return None

    @staticmethod
    def get_kernel_version() -> str:
        """Get kernel info from /proc/version"""
        with open("/proc/version", encoding="utf-8") as kernel_info:
            info = kernel_info.readlines()[0]
            version = info.split(" ")[2]
        return version

    def gamemode_available(self) -> bool:
        """Return whether gamemode is available"""
        return system.can_find_executable("gamemoderun")

    def nvidia_gamescope_support(self) -> bool:
        """Return whether gamescope is supported if we're on nvidia"""
        if not drivers.is_nvidia():
            return True

        # 515.43.04 was the first driver to support
        # VK_EXT_image_drm_format_modifier, required by gamescope.
        try:
            minimum_nvidia_version_supported = 515
            driver_info = drivers.get_nvidia_driver_info()
            driver_version = driver_info["version"]
            if not driver_version:
                return False
            major_version = int(driver_version.split(".")[0])
            return major_version >= minimum_nvidia_version_supported
        except Exception as ex:
            logger.exception("Unable to determine NVidia version: %s", ex)
            return False

    def has_steam(self) -> bool:
        """Return whether Steam is installed locally"""
        return (
            system.can_find_executable("steam")
            or flatpak.is_app_installed("com.valvesoftware.Steam")
            or os.path.exists(os.path.expanduser("~/.steam/steam/ubuntu12_32/steam"))
        )

    @property
    def display_server(self) -> str:
        """Return the display server used"""
        return os.environ.get("XDG_SESSION_TYPE", "unknown")

    def is_flatpak(self) -> bool:
        """Check is we are running inside Flatpak sandbox"""
        return system.path_exists("/.flatpak-info")

    @property
    def runtime_architectures(self) -> List[str]:
        """Return the architectures supported on this machine"""
        x86 = "i386"
        if is_exherbo_with_cross_i686():
            x86 = "libc6"
        if self.arch == "x86_64":
            return [x86, "x86_64"]
        return [x86]

    @property
    def requirements(self) -> List[str]:
        return self.get_requirements()

    @property
    def critical_requirements(self) -> List[str]:
        return self.get_requirements(include_optional=False)

    def get_fs_type_for_path(self, path: str) -> Optional[str]:
        """Return the filesystem type a given path uses"""
        mount_point = system.find_mount_point(path)
        devices = list(self.get_drives())
        while devices:
            device = devices.pop()
            devices.extend(device.get("children", []))
            if mount_point in device.get("mountpoints", []) or mount_point == device.get("mountpoint"):
                return cast(str, device["fstype"])
        return None

    def get_glxinfo(self) -> Optional[glxinfo.GlxInfo]:
        """Return a GlxInfo instance if the gfxinfo tool is available"""
        if not self.get("glxinfo"):
            return None
        _glxinfo = glxinfo.GlxInfo()
        if not hasattr(_glxinfo, "display"):
            logger.warning("Invalid glxinfo received")
            return None
        return _glxinfo

    def get_requirements(self, include_optional: bool = True) -> List[str]:
        """Return used system requirements"""
        _requirements = self.required_components.copy()
        if include_optional:
            _requirements += self.optional_components
            if drivers.is_amd():
                _requirements.append("RADEON")
        return _requirements

    def get(self, command: str) -> str:
        """Return a system command path if available"""
        return cast(str, self._cache["COMMANDS"].get(command))

    def get_terminals(self) -> List[str]:
        """Return list of installed terminals"""
        return list(self._cache["TERMINALS"].values())

    def get_soundfonts(self) -> List[str]:
        """Return path of available soundfonts"""
        return cast(List[str], self._cache["SOUNDFONTS"])

    def get_lib_folders(self) -> List[str]:
        """Return shared library folders, sorted by most used to least used"""
        lib_folder_counter = Counter(lib.dirname for lib_list in self.shared_libraries.values() for lib in lib_list)
        return [path[0] for path in lib_folder_counter.most_common()]

    def iter_lib_folders(self) -> Iterator[str]:
        """Loop over existing 32/64 bit library folders"""
        exported_lib_folders = set()
        for lib_folder in self.get_lib_folders():
            exported_lib_folders.add(lib_folder)
            yield lib_folder
        for lib_paths in self.multiarch_lib_folders:
            if self.arch != "x86_64":
                # On non amd64 setups, only the first element is relevant
                lib_paths = [lib_paths[0]]
            else:
                # Ignore paths where 64-bit path is link to supposed 32-bit path
                if os.path.realpath(lib_paths[0]) == os.path.realpath(lib_paths[1]):
                    continue
            if all(os.path.exists(path) for path in lib_paths):
                if lib_paths[0] not in exported_lib_folders:
                    yield lib_paths[0]
                if len(lib_paths) != 1:
                    if lib_paths[1] not in exported_lib_folders:
                        yield lib_paths[1]

    def get_ldconfig_libs(self) -> List[str]:
        """Return a list of available libraries, as returned by `ldconfig -p`."""
        ldconfig = self.get("ldconfig")
        if not ldconfig:
            logger.error("Could not detect ldconfig on this system")
            return []

        ld_cmd = [ldconfig, "-p"]
        if is_exherbo_with_cross_i686():
            ld_cmd = [ldconfig, "-C", "/etc/ld-i686-pc-linux-gnu.cache", "-p"]

        output = system.read_process_output(ld_cmd).split("\n")
        return [line.strip("\t") for line in output if line.startswith("\t")]

    def get_shared_libraries(self) -> Dict[str, List["SharedLibrary"]]:
        """Loads all available libraries on the system as SharedLibrary instances
        The libraries are stored in a defaultdict keyed by library name.
        """
        shared_libraries = defaultdict(list)
        for lib_line in self.get_ldconfig_libs():
            try:
                lib = SharedLibrary.new_from_ldconfig(lib_line)
            except ValueError:
                logger.error("Invalid ldconfig line: %s", lib_line)
                continue
            if lib.arch not in self.runtime_architectures:
                continue
            shared_libraries[lib.name].append(lib)
        return shared_libraries

    def populate_libraries(self) -> None:
        """Populates the LIBRARIES cache with what is found on the system"""
        self._cache["LIBRARIES"] = {}
        for arch in self.runtime_architectures:
            self._cache["LIBRARIES"][arch] = defaultdict(list)
        for req in self.requirements:
            for lib in SYSTEM_COMPONENTS["LIBRARIES"][req]:  # type: ignore
                for shared_lib in self.shared_libraries[lib]:
                    self._cache["LIBRARIES"][shared_lib.arch][req].append(lib)

    def populate_sound_fonts(self) -> None:
        """Populates the soundfont cache"""
        self._cache["SOUNDFONTS"] = []
        for folder in self.soundfont_folders:
            if not os.path.exists(folder):
                continue
            for soundfont in os.listdir(folder):
                self._cache["SOUNDFONTS"].append(soundfont)

    def get_missing_requirement_libs(self, req: str) -> List[List[str]]:
        """Return a list of sets of missing libraries for each supported architecture"""
        required_libs = set(SYSTEM_COMPONENTS["LIBRARIES"][req])  # type: ignore
        return [list(required_libs - set(self._cache["LIBRARIES"][arch][req])) for arch in self.runtime_architectures]

    def get_missing_libs(self) -> Dict[str, List[List[str]]]:
        """Return a dictionary of missing libraries"""
        return {req: self.get_missing_requirement_libs(req) for req in self.requirements}

    def get_missing_lib_arch(self, requirement: str) -> List[str]:
        """Returns a list of architectures that are missing a library for a specific
        requirement."""
        missing_arch = []
        for index, arch in enumerate(self.runtime_architectures):
            if self.get_missing_requirement_libs(requirement)[index]:
                missing_arch.append(arch)
        return missing_arch

    def is_feature_supported(self, feature: str) -> bool:
        """Return whether the system has the necessary libs to support a feature"""
        if feature == "ACO":
            try:
                mesa_version = cast(str, LINUX_SYSTEM.glxinfo.GLX_MESA_query_renderer.version)  # type: ignore
                return mesa_version >= "19.3"
            except AttributeError:
                return False
        return not self.get_missing_requirement_libs(feature)[0]

    def is_vulkan_supported(self) -> bool:
        return not LINUX_SYSTEM.get_missing_lib_arch("VULKAN") and vkquery.is_vulkan_supported()


class SharedLibrary:
    """Representation of a Linux shared library"""

    default_arch = "i386"

    def __init__(self, name: str, flags: str, path: str):
        self.name = name
        self.flags = [flag.strip() for flag in flags.split(",")]
        self.path = path

    @classmethod
    def new_from_ldconfig(cls, ldconfig_line: str) -> "SharedLibrary":
        """Create a SharedLibrary instance from an output line from ldconfig"""
        lib_match = re.match(r"^(.*) \((.*)\) => (.*)$", ldconfig_line)
        if not lib_match:
            raise ValueError("Received incorrect value for ldconfig line: %s" % ldconfig_line)
        return cls(lib_match.group(1), lib_match.group(2), lib_match.group(3))

    @property
    def arch(self) -> str:
        """Return the architecture for a shared library"""
        detected_arch = ["x86-64", "x32"]

        if is_exherbo_with_cross_i686():
            detected_arch.append("libc6")

        for arch in detected_arch:
            if arch in self.flags:
                return arch.replace("-", "_")
        return self.default_arch

    @property
    def basename(self) -> str:
        """Return the name of the library without an extention"""
        return self.name.split(".so")[0]

    @property
    def dirname(self) -> str:
        """Return the directory where the lib resides"""
        return os.path.dirname(self.path)

    def __str__(self) -> str:
        return "%s (%s)" % (self.name, self.arch)


LINUX_SYSTEM = LinuxSystem()


def gather_system_info() -> Dict[str, Any]:
    """Get all system information in a single data structure"""
    system_info = {}
    if drivers.is_nvidia():
        system_info["nvidia_driver"] = drivers.get_nvidia_driver_info()
        system_info["nvidia_gpus"] = [drivers.get_nvidia_gpu_info(gpu_id) for gpu_id in drivers.get_nvidia_gpu_ids()]
    system_info["gpus"] = [drivers.get_gpu_info(gpu) for gpu in drivers.get_gpu_cards()]
    system_info["env"] = dict(os.environ)
    system_info["missing_libs"] = LINUX_SYSTEM.get_missing_libs()
    system_info["cpus"] = LINUX_SYSTEM.get_cpus()
    system_info["drives"] = LINUX_SYSTEM.get_drives()
    system_info["ram"] = LINUX_SYSTEM.get_ram_info()
    system_info["dist"] = LINUX_SYSTEM.get_dist_info()
    system_info["arch"] = LINUX_SYSTEM.get_arch()
    system_info["kernel"] = LINUX_SYSTEM.get_kernel_version()
    system_info["glxinfo"] = glxinfo.GlxInfo().as_dict()
    return system_info


def gather_system_info_dict() -> Dict[str, Any]:
    """Get all relevant system information already formatted as a string"""
    system_info = gather_system_info()
    system_info_readable = {}
    # Add system information
    system_dict = {}
    system_dict["OS"] = " ".join(system_info["dist"])
    system_dict["Arch"] = system_info["arch"]
    system_dict["Kernel"] = system_info["kernel"]
    system_dict["Lutris Version"] = settings.VERSION
    system_dict["Python Version"] = sys.version
    system_dict["Desktop"] = system_info["env"].get("XDG_CURRENT_DESKTOP", "Not found")
    system_dict["Display Server"] = system_info["env"].get("XDG_SESSION_TYPE", "Not found")
    system_info_readable["System"] = system_dict
    # Add CPU information
    cpu_dict = {}
    cpu_dict["Vendor"] = system_info["cpus"][0].get("vendor_id", "Vendor unavailable")
    cpu_dict["Model"] = system_info["cpus"][0].get("model name", "Model unavailable")
    cpu_dict["Physical cores"] = system_info["cpus"][0].get("cpu cores", "Physical cores unavailable")
    cpu_dict["Logical cores"] = system_info["cpus"][0].get("siblings", "Logical cores unavailable")
    system_info_readable["CPU"] = cpu_dict
    # Add memory information
    ram_dict = {}
    ram_dict["RAM"] = "%0.1f GB" % (float(system_info["ram"]["MemTotal"]) / 1024 / 1024)
    ram_dict["Swap"] = "%0.1f GB" % (float(system_info["ram"]["SwapTotal"]) / 1024 / 1024)
    system_info_readable["Memory"] = ram_dict
    # Add graphics information
    graphics_dict = {}
    if LINUX_SYSTEM.glxinfo:
        graphics_dict["Vendor"] = system_info["glxinfo"].get("opengl_vendor", "Vendor unavailable")
        graphics_dict["OpenGL Renderer"] = system_info["glxinfo"].get("opengl_renderer", "OpenGL Renderer unavailable")
        graphics_dict["OpenGL Version"] = system_info["glxinfo"].get("opengl_version", "OpenGL Version unavailable")
        graphics_dict["OpenGL Core"] = system_info["glxinfo"].get(
            "opengl_core_profile_version", "OpenGL core unavailable"
        )
        graphics_dict["OpenGL ES"] = system_info["glxinfo"].get("opengl_es_profile_version", "OpenGL ES unavailable")
    else:
        graphics_dict["Vendor"] = "Unable to obtain glxinfo"
    # check Vulkan support
    if vkquery.is_vulkan_supported():
        graphics_dict["Vulkan Version"] = vkquery.format_version(vkquery.get_vulkan_api_version())

        graphics_dict["Vulkan Drivers"] = ", ".join(
            {"%s (%s)" % (name, vkquery.format_version(version)) for name, version in vkquery.get_device_info()}
        )
    else:
        graphics_dict["Vulkan"] = "Not Supported"
    system_info_readable["Graphics"] = graphics_dict
    return system_info_readable


def get_terminal_apps() -> List[str]:
    """Return the list of installed terminal emulators"""
    return LINUX_SYSTEM.get_terminals()


def get_default_terminal() -> Optional[str]:
    """Return the default terminal emulator, or None if none found."""
    terms = get_terminal_apps()
    if terms:
        return terms[0]
    return None


def get_required_default_terminal() -> str:
    """Return the default terminal emulator, or raises MisconfigurationError if none can be
    found."""
    term = get_default_terminal()
    if term:
        return term
    raise MisconfigurationError(_("No terminal emulator could be detected."))
